__authors__ = "Tim Savage"
__author_email__ = "tim@savage.company"
__copyright__ = "Copyright (C) 2014 Tim Savage"
__version__ = "1.0"

# Disable logging if an explicit handler is not added
try:
    import logging
    logging.getLogger('odin').addHandler(logging.NullHandler())
except AttributeError:
    pass  # Fallback for python 2.6

from odin.fields import *  # noqa
from odin.fields.composite import *  # noqa
from odin.fields.virtual import *  # noqa
from odin.mapping import *  # noqa
from odin.resources import Resource  # noqa
from odin.adapters import ResourceAdapter  # noqa



# -*- coding: utf-8 -*-
"""
Documentation generation from resources.
"""
import os

from odin.utils import getmeta

try:
    from jinja2 import Environment, FileSystemLoader
except ImportError:
    raise ImportError("Jinja2 is required for documentation generation support.")
from odin import registration

__all__ = ('FORMAT_TEMPLATE_RST', 'FORMAT_TEMPLATE_HTML', 'dump', 'dumps')

_TEMPLATE_ROOT = os.path.join(os.path.dirname(__file__), 'templates')

# Builtin format templates
FORMAT_TEMPLATE_RST = 'odin/doc.rst'
FORMAT_TEMPLATE_HTML = 'odin/doc.html'


class ResourceDocumentation(object):
    """
    Wrapper for resources to simplify documentation generation.
    """
    def __init__(self, resource):
        self.resource = resource
        self._meta = getmeta(resource)
        self._fields = None

    @property
    def verbose_name(self):
        return self._meta.verbose_name

    @property
    def description(self):
        return self.resource.__doc__

    @property
    def resource_name(self):
        return self._meta.resource_name

    @property
    def fields(self):
        if self._fields is None:
            self._fields = sorted(({
                "name": f.name,
                "doc_text": f.doc_text,
                "verbose_name": f.verbose_name,
                "verbose_name_plural": f.verbose_name_plural,
                "choices": [c[0] for c in f.choices or []],
                "has_default": f.has_default(),
                "default": f.get_default(),
                "optional": f.null,
                "null": f.null
            } for f in self._meta.fields), key=lambda f: f['name'])
        return self._fields


def _auto_escape(template_name):
    if template_name:
        _, ext = os.path.splitext(template_name)
        return ext in ('.html', '.xhtml', '.htm', '.xml')
    return False


def dump(fp, fmt=FORMAT_TEMPLATE_RST, exclude=None, template_path=None):
    """
    Dump resource documentation as Restructured Text.

    :param fp: File pointer to write documentation to.
    :param fmt: Format template, default is restructured text (can be used with Sphinx).
    :param exclude: List of resources to exclude from generation.
    :param template_path: An additional template_path for customisation of generated documentation.

    If an additional template path is supplied it will be made the first path in the template search paths and will
    override any built in templates.
    """
    fp.write(dumps(fmt, exclude, template_path))


def dumps(fmt=FORMAT_TEMPLATE_RST, exclude=None, template_path=None):
    """
    Dump resource documentation to a string.

    :param fmt: Format template, default is restructured text (can be used with Sphinx).
    :param exclude: List of resources to exclude from generation.
    :param template_path: An additional template_path for customisation of generated documentation.
    :returns: string representation of documentation.

    If an additional template path is supplied it will be made the first path in the template search paths and will
    override any built in templates.
    """
    exclude = exclude or []

    # Get template
    search_path = (template_path, _TEMPLATE_ROOT) if template_path else _TEMPLATE_ROOT
    env = Environment(loader=FileSystemLoader(search_path), autoescape=_auto_escape, )
    template = env.get_template(fmt)

    # Build resources list
    resources = [ResourceDocumentation(r) for r in registration.cache if getmeta(r).resource_name not in exclude]

    return template.render(resources=resources)



# -*- coding: utf-8 -*-
import odin
from odin.contrib.geo import (
    LatitudeField, LongitudeField, LatLngField, PointField,
    latitude, longitude, latlng, point
)
from odin.codecs import json_codec


class GeoResource(odin.Resource):
    class Meta:
        namespace = "odin.tests"

    lat_a = LatitudeField(null=True)
    lat_b = LatitudeField()
    lat_c = LatitudeField()
    lng_a = LongitudeField(null=True)
    lng_b = LongitudeField()
    lng_c = LongitudeField()
    latlng_a = LatLngField(null=True)
    latlng_b = LatLngField()
    point_a = PointField(null=True)
    point_b = PointField()


class TestGeoSerialisation(object):
    def test_serialise(self):
        resource = GeoResource(
            lat_a=None,
            lat_b=latitude(23.67),
            lat_c=latitude(-23.67),
            lng_a=None,
            lng_b=longitude(123.56),
            lng_c=longitude(-123.56),
            latlng_a=None,
            latlng_b=latlng(23.67, -123.56),
            point_a=None,
            point_b=point(66.66, -33.33)
        )

        actual = json_codec.dumps(resource, sort_keys=True)

        assert (
            actual ==
            '{"$": "odin.tests.GeoResource", '
            '"lat_a": null, "lat_b": 23.67, "lat_c": -23.67, '
            '"latlng_a": null, "latlng_b": [23.67, -123.56], '
            '"lng_a": null, "lng_b": 123.56, "lng_c": -123.56, '
            '"point_a": null, "point_b": [66.66, -33.33]}'
        )

    def test_deserialise(self):
        resource = json_codec.loads(
            '{"$": "odin.tests.GeoResource", '
            '"lat_a": null, "lat_b": 23.67, "lat_c": -23.67, '
            '"latlng_a": null, "latlng_b": [23.67, -123.56], '
            '"lng_a": null, "lng_b": 123.56, "lng_c": -123.56, '
            '"point_a": null, "point_b": [66.66, -33.33]}'
        )

        assert None == resource.lat_a
        assert latitude(23.67) == resource.lat_b
        assert latitude(-23.67) == resource.lat_c
        assert None == resource.lng_a
        assert longitude(123.56) == resource.lng_b
        assert longitude(-123.56) == resource.lng_c
        assert None == resource.latlng_a
        assert latlng(23.67, -123.56) == resource.latlng_b
        assert None == resource.point_a
        assert point(66.66, -33.33) == resource.point_b



import re
from odin.exceptions import ValidationError


def is_valid_ipv6_address(ip_str):
    """
    Ensure we have a valid IPv6 address.
    Args:
        ip_str: A string, the IPv6 address.
    Returns:
        A boolean, True if this is a valid IPv6 address.
    """
    # Prevent circular imports
    from odin.validators import validate_ipv4_address

    symbols_re = re.compile(r'^[0-9a-fA-F:.]+$')
    if not symbols_re.match(ip_str):
        return False

    # We need to have at least one ':'.
    if ':' not in ip_str:
        return False

    # We can only have one '::' shortener.
    if ip_str.count('::') > 1:
        return False

    # '::' should be encompassed by start, digits or end.
    if ':::' in ip_str:
        return False

    # A single colon can neither start nor end an address.
    if ((ip_str.startswith(':') and not ip_str.startswith('::')) or
            (ip_str.endswith(':') and not ip_str.endswith('::'))):
        return False

    # We can never have more than 7 ':' (1::2:3:4:5:6:7:8 is invalid)
    if ip_str.count(':') > 7:
        return False

    # If we have no concatenation, we need to have 8 fields with 7 ':'.
    if '::' not in ip_str and ip_str.count(':') != 7:
        # We might have an IPv4 mapped address.
        if ip_str.count('.') != 3:
            return False

    ip_str = _explode_shorthand_ip_string(ip_str)

    # Now that we have that all squared away, let's check that each of the
    # hextets are between 0x0 and 0xFFFF.
    for hextet in ip_str.split(':'):
        if hextet.count('.') == 3:
            # If we have an IPv4 mapped address, the IPv4 portion has to
            # be at the end of the IPv6 portion.
            if not ip_str.split(':')[-1] == hextet:
                return False
            try:
                validate_ipv4_address(hextet)
            except ValidationError:
                return False
        else:
            try:
                # a value error here means that we got a bad hextet,
                # something like 0xzzzz
                if int(hextet, 16) < 0x0 or int(hextet, 16) > 0xFFFF:
                    return False
            except ValueError:
                return False
    return True


def _explode_shorthand_ip_string(ip_str):
    """
    Expand a shortened IPv6 address.
    Args:
        ip_str: A string, the IPv6 address.
    Returns:
        A string, the expanded IPv6 address.
    """
    if not _is_shorthand_ip(ip_str):
        # We've already got a longhand ip_str.
        return ip_str

    hextet = ip_str.split('::')

    # If there is a ::, we need to expand it with zeroes
    # to get to 8 hextets - unless there is a dot in the last hextet,
    # meaning we're doing v4-mapping
    if '.' in ip_str.split(':')[-1]:
        fill_to = 7
    else:
        fill_to = 8

    if len(hextet) > 1:
        sep = len(hextet[0].split(':')) + len(hextet[1].split(':'))
        new_ip = hextet[0].split(':')

        for __ in range(fill_to - sep):
            new_ip.append('0000')
        new_ip += hextet[1].split(':')

    else:
        new_ip = ip_str.split(':')

    # Now need to make sure every hextet is 4 lower case characters.
    # If a hextet is < 4 characters, we've got missing leading 0's.
    ret_ip = []
    for hextet in new_ip:
        ret_ip.append(('0' * (4 - len(hextet)) + hextet).lower())
    return ':'.join(ret_ip)


def _is_shorthand_ip(ip_str):
    """Determine if the address is shortened.
    Args:
        ip_str: A string, the IPv6 address.
    Returns:
        A boolean, True if the address is shortened.
    """
    if ip_str.count('::') == 1:
        return True
    if any(len(x) < 4 for x in ip_str.split(':')):
        return True
    return False



# -*- coding: utf-8 -*-
import collections
import six
from odin import bases
from odin import registration
from odin.fields import NOT_PROVIDED
from odin.resources import Resource
from odin.fields.composite import ListOf, DictAs
from odin.exceptions import MappingSetupError, MappingExecutionError
from odin.mapping.helpers import MapListOf, MapDictAs, NoOpMapper
from odin.utils import cached_property, getmeta

__all__ = ('Mapping', 'map_field', 'map_list_field', 'assign_field', 'define', 'assign')


def force_tuple(value):
    return value if isinstance(value, (list, tuple)) else (value,)

EMPTY_LIST = tuple()

FieldMapping = collections.namedtuple('FieldMapping', ('from_field', 'action', 'to_field',
                                                       'to_list', 'bind', 'skip_if_none'))


def define(from_field=None, action=None, to_field=None, to_list=False, bind=False, skip_if_none=False):
    """
    Helper method for defining a mapping.

    :param from_field: Source field to map from.
    :param action: Action callable to perform during mapping, accepted fields differ based on options.
    :param to_field: Destination field to map to; if not specified the from_field
    :param to_list: Assume the result is a list (rather than a multi value tuple).
    :param bind: During the mapping operation the first parameter should be the mapping instance.
    :param skip_if_none: If the from field is :const:`None` do not include the field (this allows the destination
        object to define it's own defaults etc)
    :return: A mapping definition.

    """
    if from_field is None and to_field is None:
        raise MappingSetupError("Either `from_field` or `to_field` must be defined.")
    return FieldMapping(from_field, action, to_field or from_field, to_list, bind, skip_if_none)


def assign(to_field, action, to_list=False, bind=True, skip_if_none=False):
    """
    Helper method for defining an assignment mapping.

    :param to_field: Destination field to map to; if not specified the from_field
    :param action: Action callable to perform during mapping, accepted fields differ based on options.
    :param to_list: Assume the result is a list (rather than a multi value tuple).
    :param bind: During the mapping operation the first parameter should be the mapping instance; defaults to True
    :param skip_if_none: If the from field is :const:`None` do not include the field (this allows the destination
        object to define it's own defaults etc)
    :return: A mapping definition.

    """
    if to_field is None:
        raise MappingSetupError("`to_field` must be defined.")
    return FieldMapping(None, action, to_field, to_list, bind, skip_if_none)


class FieldResolverBase(object):
    """Base class for field resolver objects"""
    def __init__(self, obj):
        self.obj = obj

    @cached_property
    def from_field_dict(self):
        """Property accessor for the attribute dict"""
        return self.get_from_field_dict()

    def get_from_field_dict(self):
        """ Return a field map of source fields consisting of an attribute and a Field object (if one is used).

        For resource objects the field object would be an Odin resource field, for Django models a model field etc. If
        you are building a generic object mapper the field object can be :const:`None`.

        The field object is used to determine certain automatic mapping operations (ie Lists of Odin resources to other
        Odin resources).

        :return: Dictionary

        """
        return self.get_field_dict()

    @cached_property
    def to_field_dict(self):
        """Property accessor for the attribute dict"""
        return self.get_to_field_dict()

    def get_to_field_dict(self):
        """ Return a field map consisting of an attribute and a Field object (if one is used).

        For resource objects the field object would be an Odin resource field, for Django models a model field etc. If
        you are building a generic object mapper the field object can be :const:`None`.

        The field object is used to determine certain automatic mapping operations (ie Lists of Odin resources to other
        Odin resources).

        :return: Dictionary
        """
        return self.get_field_dict()

    def get_field_dict(self):
        """ Return a field map consisting of an attribute and a Field object (if one is used).

        For resource objects the field object would be an Odin resource field, for Django models a model field etc. If
        you are building a generic object mapper the field object can be :const:`None`.

        The field object is used to determine certain automatic mapping operations (ie Lists of Odin resources to other
        Odin resources).

        :return: Dictionary
        """
        raise NotImplementedError()


class ResourceFieldResolver(FieldResolverBase):
    """Field resolver for Odin resource objects."""
    def get_field_dict(self):
        """Return a dictionary of fields along with their names."""
        return getmeta(self.obj).field_map

registration.register_field_resolver(ResourceFieldResolver, Resource)


class MappingMeta(type):
    """
    Meta-class for all Mappings
    """
    def __new__(mcs, name, bases, attrs):
        super_new = super(MappingMeta, mcs).__new__

        # attrs will never be empty for classes declared in the standard way
        # (ie. with the `class` keyword). This is quite robust.
        if name == 'NewBase' and attrs == {}:
            return super_new(mcs, name, bases, attrs)

        parents = [
            b for b in bases
            if isinstance(b, MappingMeta) and not (b.__name__ == 'NewBase' and b.__mro__ == (b, MappingBase, object))
        ]
        if not parents:
            # If this isn't a subclass of Mapping, don't do anything special.
            return super_new(mcs, name, bases, attrs)

        # Backward compatibility from_resource -> from_obj
        from_obj = attrs.setdefault('from_obj', attrs.get('from_resource'))
        if from_obj is None:
            raise MappingSetupError('`from_obj` is not defined.')
        to_obj = attrs.setdefault('to_obj', attrs.get('to_resource'))
        if to_obj is None:
            raise MappingSetupError('`to_obj` is not defined.')
        register_mapping = attrs.pop('register_mapping', True)

        # Check if we have already created this mapping
        if register_mapping:
            try:
                return registration.get_mapping(from_obj, to_obj)
            except KeyError:
                pass  # Not registered

        # Get field resolver objects
        try:
            from_fields = registration.get_field_resolver(from_obj).from_field_dict
        except KeyError:
            raise MappingSetupError('`from_obj` %r does not have an attribute resolver defined.' % from_obj)
        try:
            to_fields = registration.get_field_resolver(to_obj).to_field_dict
        except KeyError:
            raise MappingSetupError('`to_obj` %r does not have an attribute resolver defined.' % to_obj)

        def attr_mapping_to_mapping_rule(m, def_type, ref):
            """ Parse, validate and normalise defined mapping rules so rules can be executed without having to perform
            checks during a mapping operation."""
            to_list = False
            bind = False
            skip_if_none = False
            is_assignment = False
            try:
                map_from, action, map_to, to_list, bind, skip_if_none = m
            except ValueError:
                try:
                    map_from, action, map_to = m
                except ValueError:
                    raise MappingSetupError('Bad mapping definition `%s` in %s `%s`.' % (m, def_type, ref))

            if map_from is None:
                is_assignment = True

            if not is_assignment:
                map_from = force_tuple(map_from)
                for f in map_from:
                    if f not in from_fields:
                        raise MappingSetupError('Field `%s` of %s `%s` not found on from object. ' % (f, def_type, ref))

            if isinstance(action, six.string_types):
                if action not in attrs:
                    raise MappingSetupError('Action named %s defined in %s `%s` was not defined on mapping object.' % (
                        action, def_type, ref))
                if not callable(attrs[action]):
                    raise MappingSetupError('Action named %s defined in %s `%s` is not callable.' % (
                        action, def_type, ref))
            elif action is not None and not callable(action):
                raise MappingSetupError('Action on %s `%s` is not callable.' % (def_type, ref))
            elif action is None and is_assignment:
                raise MappingSetupError('No action supplied for `%s` in `%s`.' % (def_type, ref))

            map_to = force_tuple(map_to)
            if to_list and len(map_to) != 1:
                raise MappingSetupError('The %s `%s` specifies a to_list mapping, these can only be applied to a '
                                        'single target field.' % (def_type, m))
            for f in map_to:
                if f not in to_fields:
                    raise MappingSetupError('Field `%s` of %s `%s` not found on to object. ' % (f, def_type, ref))

            return FieldMapping(map_from, action, map_to, to_list, bind, skip_if_none)

        # Determine what fields need to have mappings generated
        exclude_fields = attrs.get('exclude_fields') or tuple()
        unmapped_fields = [attname for attname in from_fields if attname not in exclude_fields]

        def remove_from_unmapped_fields(rule):
            # Remove any fields that are handled by a mapping rule from unmapped_fields list.
            map_to = rule[2]
            if len(map_to) == 1 and map_to[0] in unmapped_fields:
                unmapped_fields.remove(map_to[0])

        # Generate mapping rules.
        mapping_rules = []

        # Check that from_obj is a sub_class (or same class) as any `parent.from_obj`. This is important for mapping
        # sub class lists and resolving mappings.
        base_parents = [p for p in parents if hasattr(p, '_subs')]
        for p in base_parents:
            if not issubclass(from_obj, p.from_obj):
                raise MappingSetupError('`from_obj` must be a subclass of `parent.from_obj`')
            if not issubclass(to_obj, p.to_obj):
                raise MappingSetupError('`to_obj` must be a subclass of `parent.to_obj`')

            # Copy mapping rules
            for mapping_rule in p._mapping_rules:
                mapping_rules.append(mapping_rule)
                remove_from_unmapped_fields(mapping_rule)

        # Add basic mappings
        for idx, mapping in enumerate(attrs.pop('mappings', [])):
            mapping_rule = attr_mapping_to_mapping_rule(mapping, 'basic mapping', idx)
            mapping_rules.append(mapping_rule)
            remove_from_unmapped_fields(mapping_rule)

        # Add custom mappings
        for attr in attrs.values():
            # Methods with a _mapping attribute have been decorated by either `map_field` or `map_list_field`
            # decorators.
            if hasattr(attr, '_mapping'):
                mapping_rule = attr_mapping_to_mapping_rule(getattr(attr, '_mapping'), 'custom mapping', attr)
                mapping_rules.append(mapping_rule)
                remove_from_unmapped_fields(mapping_rule)
                # Remove mapping
                delattr(attr, '_mapping')

        # Add auto mapped fields that are yet to be mapped.
        for field in unmapped_fields:
            if field in to_fields:
                mapping_rules.append(mcs.generate_auto_mapping(field, from_fields, to_fields))

        # Update attributes
        attrs['_mapping_rules'] = mapping_rules
        attrs['_subs'] = {}

        # Create mapper instance
        mapper = super_new(mcs, name, bases, attrs)
        if register_mapping:
            registration.register_mapping(mapper)
            mapper = registration.get_mapping(from_obj, to_obj)

            # Register mapping with parents mapping objects as a sub class.
            for parent in base_parents:
                parent._subs[from_obj] = mapper

        return mapper

    @classmethod
    def generate_auto_mapping(mcs, name, from_fields, to_fields):
        """
        Generate the auto mapping between two fields.
        """
        from_field = from_fields[name]
        to_field = to_fields[name]
        name = force_tuple(name)

        # Handle ListOf fields
        if isinstance(from_field, ListOf) and isinstance(to_field, ListOf):
            return mcs.generate_list_to_list_mapping(name, from_field, to_field)

        # Handle DictAs fields
        elif isinstance(from_field, DictAs) and isinstance(to_field, DictAs):
            return mcs.generate_dict_to_dict_mapping(name, from_field, to_field)

        return define(name, 'default_action')

    @classmethod
    def generate_list_to_list_mapping(mcs, name, from_field, to_field):
        """
        Generate a mapping of list to list objects.
        """
        try:
            mapping = registration.get_mapping(from_field.of, to_field.of)
            return FieldMapping(name, MapListOf(mapping), name, to_list=False, bind=True, skip_if_none=False)
        except KeyError:
            # If both items are from and to fields refer to the same object automatically use a mapper that just
            # produces a clone.
            if from_field.of is to_field.of:
                return FieldMapping(name, MapListOf(NoOpMapper), name, to_list=False, bind=True, skip_if_none=False)

    @classmethod
    def generate_dict_to_dict_mapping(mcs, name, from_field, to_field):
        try:
            mapping = registration.get_mapping(from_field.of, to_field.of)
            return define(name, MapDictAs(mapping), name, bind=True)
        except KeyError:
            # If both items are from and to fields refer to the same object automatically use a mapper that just
            # produces a clone.
            if from_field.of is to_field.of:
                return define(name, MapDictAs(NoOpMapper), name, bind=True)


class MappingResult(bases.TypedResourceIterable):
    """
    Iterator used lazily return a sequence from a mapping operation (used by ``Mapping.apply``).
    """
    def __init__(self, sequence, mapping, context=None, *mapping_options):
        super(MappingResult, self).__init__(mapping.to_obj)
        self.sequence = sequence
        self.mapping = mapping
        self.context = context or {}
        self.context.setdefault('_loop_idx', [])
        self.mapping_options = mapping_options

    def __iter__(self):
        self.context['_loop_idx'].append(0)
        for item in self.sequence:
            yield self.mapping.apply(item, self.context, *self.mapping_options)
            self.context['_loop_idx'][-1] += 1
        self.context['_loop_idx'].pop()


class ImmediateResult(bases.ResourceIterable):
    """
    Immediately performs the mapping operation rather than delay.

    This is useful if context is volatile.
    """
    def __init__(self, *args, **kwargs):
        self.sequence = list(MappingResult(*args, **kwargs))

    def __iter__(self):
        for item in self.sequence:
            yield item


class CachingMappingResult(MappingResult):
    """
    Extends from the basic MappingResult to cache the results of a mapping operation and also provide methods and
    abilities available to a list (len, indexing). The results are only evaluated when requested.
    """
    def __init__(self, *args, **kwargs):
        super(CachingMappingResult, self).__init__(*args, **kwargs)
        self._cache = None  # type: list

    def __iter__(self):
        if self._cache is None:
            self._cache = []
            for item in super(CachingMappingResult, self).__iter__():
                self._cache.append(item)
                yield item
        else:
            # Python 2.x compatible (python 3 can use yield from)
            for item in self._cache:
                yield item

    @property
    def items(self):
        # type: () -> list
        if self._cache is None:
            list(iter(self))
        return self._cache

    def __len__(self):
        return len(self.items)

    def __getitem__(self, idx):
        return self.items[idx]


class MappingBase(object):
    from_obj = None  # type: type
    to_obj = None  # type: type

    # Pending deprecation, move to from_obj and to_obj terminology
    from_resource = None
    to_resource = None

    # Default mapping result object to use
    default_mapping_result = CachingMappingResult

    # Register mapping with global registry, if the same resource pair
    # is mapped in different ways then this might want to be set to False
    register_mapping = True

    _mapping_rules = None

    @classmethod
    def apply(cls, source_obj, context=None, allow_subclass=False, mapping_result=None):
        """
        Apply conversion either a single resource or a list of resources using the mapping defined by this class.

        If a list of resources is supplied an iterable is returned.

        :param source_obj: The source resource, this must be an instance of :py:attr:`Mapping.from_obj`.
        :param context: An optional context value, this can be any value you want to aid in mapping
        :param allow_subclass: Allow sub classes of mapping resource to be included.
        :param mapping_result: If an iterable is provided as the source object a mapping result is returned of the type
            specified is returned.

        """
        context = context or {}
        mapping_result = mapping_result or cls.default_mapping_result
        context.setdefault('_loop_idx', [])

        if hasattr(source_obj, '__iter__'):
            return mapping_result(source_obj, cls, context, allow_subclass)
        elif source_obj.__class__ is cls.from_obj:
            return cls(source_obj, context).convert()
        else:
            # Sub class lookup required
            sub_mapping = cls._subs.get(source_obj.__class__)
            if sub_mapping:
                return sub_mapping(source_obj, context).convert()
            if allow_subclass:
                if allow_subclass and isinstance(source_obj, cls.from_obj):
                    return cls(source_obj, context, True).convert()

                raise TypeError('`source_resource` parameter must be an instance (or subclass instance) of %s' %
                                cls.from_obj)

            raise TypeError('`source_resource` parameter must be an instance of %s' % cls.from_obj)

    def __init__(self, source_obj, context=None, allow_subclass=False):
        """
        Initialise instance of mapping.

        :param source_obj: The source resource, this must be an instance of :py:attr:`Mapping.from_obj`.
        :param context: An optional context value, this can be any value you want to aid in mapping
        """
        if allow_subclass:
            if not isinstance(source_obj, self.from_obj):
                raise TypeError('`source_resource` parameter must be an instance of subclass of %s' % self.from_obj)
        else:
            if source_obj.__class__ is not self.from_obj:
                raise TypeError('`source_resource` parameter must be an instance of %s' % self.from_obj)
        self.source = source_obj
        self.context = context or {}

    @property
    def loop_idx(self):
        """
        Index within a loop of this mapping (note loop might be for a parent object)
        :returns: Index within the loop; or `None` if we are not currently in a loop.
        """
        loops = self.context.setdefault('_loop_idx', [])
        return loops[-1] if loops else None

    @property
    def loop_level(self):
        """
        How many layers of loops are we in?
        """
        return len(self.context.setdefault('_loop_idx', []))

    @property
    def in_loop(self):
        """
        Is this mapping currently in a loop?
        """
        return bool(self.context.setdefault('_loop_idx', []))

    def default_action(self, value):
        """
        The default action used when mapping. This is a bit of a special case in that it defaults to being bound
        and makes use of of :func:`functools.partial` to bind the from and to fields.

        :param value: The value to be mapped.
        :return: Mapped value.

        """
        """
        The default action that is applied for automatic mappings.
        """
        return value

    def _apply_rule(self, mapping_rule):
        # Unpack mapping definition and fetch from values
        from_fields, action, to_fields, to_list, bind, skip_if_none = mapping_rule

        # This is an assignment rather than a mapping
        if from_fields is None:
            from_values = EMPTY_LIST
        else:
            from_values = tuple(getattr(self.source, f) for f in from_fields)

        if action is None:
            to_values = from_values
        else:
            if isinstance(action, six.string_types):
                action = getattr(self, action)

            try:
                if bind:
                    to_values = action(self, *from_values)
                else:
                    to_values = action(*from_values)
            except TypeError as ex:
                raise MappingExecutionError('%s applying rule %s' % (ex, mapping_rule))

        if to_list:
            if isinstance(to_values, collections.Iterable):
                to_values = (list(to_values),)
        else:
            to_values = force_tuple(to_values)

        if len(to_fields) != len(to_values):
            raise MappingExecutionError('Rule expects %s fields (%s returned) applying rule %s' % (
                len(to_fields), len(to_values), mapping_rule))

        if skip_if_none:
            return dict((f, to_values[i]) for i, f in enumerate(to_fields) if to_values[i] is not None)
        else:
            return dict((f, to_values[i]) for i, f in enumerate(to_fields))

    def create_object(self, **field_values):
        """
        Create an instance of target object, this method can be customise to handle custom object initialisation.

        :param field_values: Dictionary of values for creating the target object.

        """
        return self.to_obj(**field_values)

    def convert(self, **field_values):
        """
        Convert the provided source into a destination object.

        :param field_values: Initial field values (or fields not provided by source object);

        """
        assert hasattr(self, '_mapping_rules')

        values = field_values

        for mapping_rule in self._mapping_rules:
            values.update(self._apply_rule(mapping_rule))

        return self.create_object(**values)

    def update(self, destination_obj, ignore_fields=None, fields=None):
        """
        Update an existing object with fields from the provided source object.

        :param destination_obj: The existing destination object.
        :param ignore_fields: A list of fields that should be ignored eg ID fields
        :param fields: Collection of fields that should be mapped.

        """
        assert hasattr(self, '_mapping_rules')

        ignore_fields = ignore_fields or []

        for mapping_rule in self._mapping_rules:
            for name, value in self._apply_rule(mapping_rule).items():
                if name in ignore_fields or (fields and name not in fields) or value is NOT_PROVIDED:
                    continue
                setattr(destination_obj, name, value)

        return destination_obj

    def diff(self, destination_obj):
        """
        Return all fields that are different.

        :note: a full mapping operation is performed during the diffing process.

        :param destination_obj: The existing destination object.
        :return: set of fields that vary.

        """
        assert hasattr(self, '_mapping_rules')

        diff_fields = set()
        for mapping_rule in self._mapping_rules:
            for name, value in self._apply_rule(mapping_rule).items():
                if value != getattr(destination_obj, name):
                    diff_fields.add(name)
        return diff_fields


class Mapping(six.with_metaclass(MappingMeta, MappingBase)):
    """
    Definition of a mapping between two Objects.
    """
    exclude_fields = []
    mappings = []


class DynamicMapping(MappingBase):
    """
    A mapping that is dynamically generated at run time.
    """


class ImmediateMapping(six.with_metaclass(MappingMeta, MappingBase)):
    """
    Definition of a mapping between two Objects.

    This version of the Mapping, defaults to returning an Immediate rather than
    a cached result object.

    """
    exclude_fields = []
    mappings = []

    default_mapping_result = ImmediateResult


def map_field(func=None, from_field=None, to_field=None, to_list=False):
    """
    Field decorator for custom mappings.

    :param func: Method being decorator is wrapping.
    :param from_field: Name of the field to map from; default is to use the function name.
    :param to_field: Name of the field to map to; default is to use the function name.
    :param to_list: The result is a list (rather than a multi value tuple).
    """
    def inner(fun):
        fun._mapping = define(
            from_field or fun.__name__,
            fun.__name__,
            to_field or fun.__name__,
            to_list
        )
        return fun

    return inner(func) if func else inner


def map_list_field(*args, **kwargs):
    """
    Field decorator for custom mappings that return a single list.

    This mapper also allows for returning an iterator or generator that will be converted into a list during the
    mapping operation.

    Parameters are identical to the :py:meth:`map_field` method except ``to_list`` which is forced to be True.
    """
    kwargs['to_list'] = True
    return map_field(*args, **kwargs)


def assign_field(func=None, to_field=None, to_list=False):
    """
    Field decorator for assigning a value to destination field without requiring a corresponding source field.

    Allows for the mapping to calculate a value based on the context or other information. Useful when a destination
    objects defaulting mechanism is not able to calculate a default that either applies or is suitable.

    :param func: Method being decorator is wrapping.
    :param to_field: Name of the field to assign value to; default is to use the function name.
    :param to_list: The result is a list (rather than a multi value tuple).
    """
    def inner(fun):
        fun._mapping = define(
            None,
            fun.__name__,
            to_field or fun.__name__,
            to_list
        )
        return fun

    return inner(func) if func else inner


def mapping_factory(from_obj, to_obj, base_mapping=Mapping,
                    generate_reverse=True,
                    mappings=None, reverse_mappings=None,
                    exclude_fields=None, reverse_exclude_fields=None,
                    register_mappings=True):
    """
    Factory method for generating simple mappings between objects.

    A common use-case for this method is in generating mappings in baldr's ``model_resource_factory`` method that
    auto-generates resources from Django models.

    :param from_obj: Object to map from.
    :param to_obj: Object to map to.
    :param base_mapping: Base mapping class; default is ``odin.Mapping``.
    :param generate_reverse: Generate the reverse of the mapping ie swap from_obj and to_obj.
    :param mappings: User provided mappings (this is equivalent ot ``odin.Mapping.mappings``)
    :param reverse_mappings: User provided reverse mappings (this is equivalent ot ``odin.Mapping.mappings``). Only
        used if ``generate_reverse`` is True.
    :param exclude_fields: Fields to exclude from auto-generated mappings
    :param reverse_exclude_fields: Fields to exclude from auto-generated reverse mappings.
    :param register_mappings: Register mapping in mapping lookup cache.
    :return: if generate_reverse is True a tuple(forward_mapping, reverse_mapping); else just the forward_mapping.
    :rtype: Mapping | (Mapping, Mapping)

    """
    forward_mapping = type(
        "%s%sTo%s%s" % (
            from_obj.__class__.__name__, from_obj.__name__,
            to_obj.__class__.__name__, to_obj.__name__
        ),
        (base_mapping, ),
        dict(
            from_obj=from_obj,
            to_obj=to_obj,
            exclude_fields=exclude_fields or list(),
            mappings=mappings or dict(),
            register_mapping=register_mappings
        )
    )

    if generate_reverse:
        reverse_mapping = type(
            "%s%sTo%s%s" % (
                to_obj.__class__.__name__, to_obj.__name__,
                from_obj.__class__.__name__, from_obj.__name__,
            ),
            (base_mapping, ),
            dict(
                from_obj=to_obj,
                to_obj=from_obj,
                exclude_fields=reverse_exclude_fields or list(),
                mappings=reverse_mappings or dict(),
                register_mapping=register_mappings
            )
        )

        return forward_mapping, reverse_mapping

    return forward_mapping


def forward_mapping_factory(from_obj, to_obj, base_mapping=Mapping, mappings=None, exclude_fields=None):
    """
    Factory method for generating simple forward mappings between objects.

    :param from_obj: Object to map from.
    :param to_obj: Object to map to.
    :param base_mapping: Base mapping class; default is ``odin.Mapping``.
    :param mappings: User provided mappings (this is equivalent ot ``odin.Mapping.mappings``)
    :param exclude_fields: Fields to exclude from auto-generated mappings
    :return: Forward_mapping object.

    """
    return mapping_factory(from_obj, to_obj, base_mapping, False, mappings, None, exclude_fields, None)


def dynamic_mapping_factory(from_obj, to_obj, base_mapping=Mapping, generate_reverse=False,
                            mappings=None, reverse_mappings=None,
                            exclude_fields=None, reverse_exclude_fields=None):
    """
    Factory method for generating a dynamic mapping. That is generated dynamically at run time (eg from
    configuration the results of which are not registered for later use.

    :param from_obj: Object to map from.
    :param to_obj: Object to map to.
    :param base_mapping: Base mapping class; default is ``odin.Mapping``.
    :param generate_reverse: Generate the reverse of the mapping ie swap from_obj and to_obj.
    :param mappings: User provided mappings (this is equivalent ot ``odin.Mapping.mappings``)
    :param reverse_mappings: User provided reverse mappings (this is equivalent ot ``odin.Mapping.mappings``). Only
        used if ``generate_reverse`` is True.
    :param exclude_fields: Fields to exclude from auto-generated mappings
    :param reverse_exclude_fields: Fields to exclude from auto-generated reverse mappings.
    :return: if generate_reverse is True a tuple(forward_mapping, reverse_mapping); else just the forward_mapping.
    :rtype: Mapping | (Mapping, Mapping)

    """
    return mapping_factory(from_obj, to_obj, base_mapping, generate_reverse, mappings, reverse_mappings,
                           exclude_fields, reverse_exclude_fields, False)



import datetime
import odin.datetimeutil

ARE_YOU_EXPERIENCED = datetime.date(1967, 5, 12)
MWT = odin.datetimeutil.FixedTimezone(-6, 'Mountain War Time')
BOOM = datetime.datetime(1945, 7, 16, 5, 29, 45, 0, MWT)



# -*- coding: utf-8 -*-
from copy import deepcopy
import pytest
import datetime
from odin.fields import *
from odin.fields import Field, TimeStampField, NOT_PROVIDED
from odin.datetimeutil import utc, FixedTimezone
from odin.fields.virtual import VirtualField
from odin.validators import MinValueValidator, MaxValueValidator, MaxLengthValidator, RegexValidator
from odin.exceptions import ValidationError


class ObjectValue(object):
    pass


class ValidatorTest(object):
    message = 'Default message'
    code = 'test_code'

    def __init__(self, fail, *params):
        self.fail = fail
        self.params = params

    def __call__(self, value):
        if self.fail:
            raise ValidationError(code=self.code, message=self.message, params=self.params)


class FieldTest(Field):
    def to_python(self, value):
        return value


class DynamicTypeNameFieldTest(IntegerField):
    @staticmethod
    def data_type_name(instance):
        return "Foo"


class TestField(object):
    def test_error_messages_no_overrides(self):
        target = FieldTest()

        assert {
                   'invalid_choice': 'Value %r is not a valid choice.',
                   'null': 'This field cannot be null.',
                   'required': 'This field is required.',
               } == target.error_messages

    def test_error_messages_override_add(self):
        target = FieldTest(error_messages={
            'null': 'Override',
            'other': 'Other Value',
        })

        assert {
                   'invalid_choice': 'Value %r is not a valid choice.',
                   'null': 'Override',
                   'required': 'This field is required.',
                   'other': 'Other Value',
               } == target.error_messages

    def test_set_attributes_from_name(self):
        target = FieldTest()
        target.set_attributes_from_name("test_name")

        assert "test_name" == target.name
        assert "test_name" == target.attname
        assert "test name" == target.verbose_name
        assert "test names" == target.verbose_name_plural

    def test_set_attributes_from_name_with_name(self):
        target = FieldTest(name="init_name")
        target.set_attributes_from_name("test_name")

        assert "init_name" == target.name
        assert "test_name" == target.attname
        assert "init name" == target.verbose_name
        assert "init names" == target.verbose_name_plural

    def test_set_attributes_from_name_with_verbose_name(self):
        target = FieldTest(verbose_name="init Verbose Name")
        target.set_attributes_from_name("test_name")

        assert "test_name" == target.name
        assert "test_name" == target.attname
        assert "init Verbose Name" == target.verbose_name
        assert "init Verbose Names" == target.verbose_name_plural

    def test_has_default(self):
        target = FieldTest()

        assert not target.has_default()

    def test_has_default_supplied(self):
        target = FieldTest(default="123")

        assert target.has_default()

    def test_get_default(self):
        target = FieldTest()

        assert target.get_default() is None

    def test_get_default_supplied(self):
        target = FieldTest(default="123")

        assert "123" == target.get_default()

    def test_get_default_callable(self):
        target = FieldTest(default=lambda: "321")

        assert "321" == target.get_default()

    def test_value_from_object(self):
        target = FieldTest()
        target.set_attributes_from_name("test_name")

        an_obj = ObjectValue()
        setattr(an_obj, "test_name", "test_value")

        actual = target.value_from_object(an_obj)
        assert "test_value" == actual

    def test__repr(self):
        target = Field()
        assert "<odin.fields.Field>" == repr(target)
        target.set_attributes_from_name("eek")
        assert "<odin.fields.Field: eek>" == repr(target)

    def test__deep_copy(self):
        field = FieldTest(name="Test")
        target_copy = deepcopy(field)
        target_assign = field
        assert field is target_assign
        assert field is not target_copy
        assert field.name == target_copy.name

    def test_run_validators_and_override_validator_message(self):
        target = FieldTest(error_messages={'test_code': 'Override message'}, validators=[ValidatorTest(True)])

        try:
            target.run_validators("Placeholder")
        except ValidationError as ve:
            assert 'Override message' == ve.messages[0]
        else:
            raise AssertionError("Validation Error not raised.")

    def test_run_validators_and_override_validator_message_with_params(self):
        target = FieldTest(error_messages={'test_code': 'Override message: %s'},
                           validators=[ValidatorTest(True, "123")])

        try:
            target.run_validators("Placeholder")
        except ValidationError as ve:
            assert ['Override message: 123'] == ve.messages
        else:
            raise AssertionError("Validation Error not raised.")

    def test_default_to_python_raises_not_implemented(self):
        target = Field()
        pytest.raises(NotImplementedError, target.to_python, "Anything...")

    def test_clean_uses_default_if_value_is_not_provided_is_true(self):
        target = FieldTest(use_default_if_not_provided=True, default='foo')
        actual = target.clean(NOT_PROVIDED)
        assert 'foo' == actual

    def test_clean_uses_default_if_value_is_not_provided_is_false(self):
        # Need to allow None as the if use_default_if_not_provided is false NOT_PROVIDED evaluates to None.
        target = FieldTest(use_default_if_not_provided=False, default='foo', null=True)
        actual = target.clean(NOT_PROVIDED)
        assert actual is None


class VirtualFieldTest(VirtualField):
    pass


class TestVirtualField(object):
    def test_creation_counter(self):
        current_count = VirtualField.creation_counter
        next_count = current_count + 1
        target = VirtualField()

        assert VirtualField.creation_counter == next_count
        assert target.creation_counter == current_count
        assert hash(target) == current_count

    def test_repr(self):
        target = VirtualFieldTest()
        assert "<tests.test_fields.VirtualFieldTest>" == repr(target)
        target.set_attributes_from_name("eek")
        assert "<tests.test_fields.VirtualFieldTest: eek>" == repr(target)

    def test_default_descriptor_behaviour(self):
        class TestObj(object):
            test_field = VirtualField()

        target = TestObj()

        with pytest.raises(NotImplementedError):
            _ = target.test_field

        with pytest.raises(AttributeError) as excinfo:
            target.test_field = 123

        assert "Read only" == str(excinfo.value)


class TestFields(object):
    def assertValidatorIn(self, validatorClass, validators):
        """
        Assert that the specified validator is in the validation list.
        :param validatorClass:
        :param validators:
        """
        for v in validators:
            if isinstance(v, validatorClass):
                return
        raise AssertionError("Validator %r was not found in list of validators." % validatorClass)

    def assertValidatorNotIn(self, validatorClass, validators):
        """
        Assert that the specified validator is not in the validation list.
        :param validatorClass:
        :param validators:
        """
        for v in validators:
            if isinstance(v, validatorClass):
                raise AssertionError("Validator %r was found in list of validators." % validatorClass)

    # BooleanField ############################################################

    @pytest.mark.parametrize(('field', 'value', 'actual'), (
            (BooleanField(), True, True),
            (BooleanField(), 1, True),
            (BooleanField(), 'Yes', True),
            (BooleanField(), 'true', True),
            (BooleanField(), 'T', True),
            (BooleanField(), '1', True),
            (BooleanField(), 'TRUE', True),
            (BooleanField(), False, False),
            (BooleanField(), 0, False),
            (BooleanField(), 'No', False),
            (BooleanField(), 'false', False),
            (BooleanField(), 'FALSE', False),
            (BooleanField(), 'F', False),
            (BooleanField(), '0', False),
            (BooleanField(null=True), None, None),
    ))
    def test_booleanfield_success(self, field, value, actual):
        assert field.clean(value) == actual

    @pytest.mark.parametrize(('field', 'value'), (
            (BooleanField(), None),
            (BooleanField(), ''),
            (BooleanField(), 'Awesome!'),
    ))
    def test_booleanfield_failure(self, field, value):
        with pytest.raises(ValidationError):
            field.clean(value)

    # StringField #############################################################

    @pytest.mark.parametrize(('field', 'value', 'actual'), (
            (StringField(), '1', '1'),
            (StringField(), 'eek', 'eek'),
            (StringField(null=True), '1', '1'),
            (StringField(null=True), None, None),
            (StringField(null=True), '', ''),
            (StringField(null=True, empty=True), '', ''),
            (StringField(empty=True), '', ''),
            (StringField(max_length=10), '123456', '123456'),
    ))
    def test_stringfield_success(self, field, value, actual):
        assert field.clean(value) == actual

    @pytest.mark.parametrize(('field', 'value'), (
            (StringField(), None),
            (StringField(), ''),
            (StringField(null=True, empty=False), ''),
            (StringField(empty=False), ''),
            (StringField(max_length=10), '1234567890a'),
    ))
    def test_stringfield_failure(self, field, value):
        with pytest.raises(ValidationError):
            field.clean(value)

    def test_stringfield(self):
        f = StringField()
        assert f.max_length is None
        self.assertValidatorNotIn(MaxLengthValidator, f.validators)

        f = StringField(max_length=10)
        assert f.max_length == 10
        self.assertValidatorIn(MaxLengthValidator, f.validators)

    @pytest.mark.parametrize(('field', 'empty_value'), (
            (StringField(), False),
            (StringField(null=True), True),
            (StringField(null=False), False),
            (StringField(null=True, empty=True), True),
            (StringField(null=False, empty=True), True),
            (StringField(null=True, empty=False), False),
            (StringField(null=False, empty=False), False),
    ))
    def test_stringfield__handling_of_null_empty(self, field, empty_value):
        assert field.empty == empty_value

    # URLField ################################################################

    def test_urlfield_1(self):
        f = UrlField()
        assert 'http://www.github.com' == f.clean('http://www.github.com')
        pytest.raises(ValidationError, f.clean, 'eek')
        pytest.raises(ValidationError, f.clean, None)
        assert f.max_length == None
        self.assertValidatorIn(RegexValidator, f.validators)

    # IntegerField ############################################################

    def test_integerfield_1(self):
        f = IntegerField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        assert 123 == f.clean(123)
        assert 123 == f.clean('123')
        assert 123 == f.clean(123.5)
        assert None == f.min_value
        self.assertValidatorNotIn(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assertValidatorNotIn(MaxValueValidator, f.validators)

    def test_integerfield_2(self):
        f = IntegerField(null=True)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, 'abc')
        assert 69 == f.clean(69)
        assert 69 == f.clean('69')
        assert 69 == f.clean(69.5)
        assert None == f.min_value
        self.assertValidatorNotIn(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assertValidatorNotIn(MaxValueValidator, f.validators)

    def test_integerfield_3(self):
        f = IntegerField(min_value=50, max_value=100)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        assert 69 == f.clean(69)
        assert 69 == f.clean('69')
        assert 69 == f.clean(69.5)
        assert 50 == f.clean(50)
        assert 100 == f.clean(100)
        pytest.raises(ValidationError, f.clean, 30)
        pytest.raises(ValidationError, f.clean, 110)
        assert 50 == f.min_value
        self.assertValidatorIn(MinValueValidator, f.validators)
        assert 100 == f.max_value
        self.assertValidatorIn(MaxValueValidator, f.validators)

    # FloatField ##############################################################

    def test_floatfield_1(self):
        f = FloatField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        assert 123 == f.clean(123)
        assert 123.5 == f.clean('123.5')
        assert 123.5 == f.clean(123.5)
        assert None == f.min_value
        self.assertValidatorNotIn(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assertValidatorNotIn(MaxValueValidator, f.validators)

    def test_floatfield_2(self):
        f = FloatField(null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, 'abc')
        assert 69 == f.clean(69)
        assert 69.5 == f.clean('69.5')
        assert 69.5 == f.clean(69.5)
        assert None == f.min_value
        self.assertValidatorNotIn(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assertValidatorNotIn(MaxValueValidator, f.validators)

    def test_floatfield_3(self):
        f = FloatField(min_value=50.5, max_value=100.4)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        assert 69 == f.clean(69)
        assert 69.5 == f.clean('69.5')
        assert 69.5 == f.clean(69.5)
        assert 50.5 == f.clean(50.5)
        assert 100.4 == f.clean(100.4)
        pytest.raises(ValidationError, f.clean, 30)
        pytest.raises(ValidationError, f.clean, 110)
        assert 50.5 == f.min_value
        self.assertValidatorIn(MinValueValidator, f.validators)
        assert 100.4 == f.max_value
        self.assertValidatorIn(MaxValueValidator, f.validators)

    # DateField ###############################################################

    @pytest.mark.parametrize('field,value,expected'.split(','), (
            (DateField(), '2013-11-24', datetime.date(2013, 11, 24)),
            (DateField(), datetime.date(2013, 11, 24), datetime.date(2013, 11, 24)),
            (DateField(), datetime.datetime(2013, 11, 24, 1, 14), datetime.date(2013, 11, 24)),
            (DateField(null=True), None, None),
    ))
    def test_datefield_clean_success(self, field, value, expected):
        assert field.clean(value) == expected

    @pytest.mark.parametrize('field,value'.split(','), (
            (DateField(), None),
            (DateField(), 'abc'),
            (DateField(), 123),
    ))
    def test_datefield_clean_failure(self, field, value):
        pytest.raises(ValidationError, field.clean, value)

    @pytest.mark.parametrize('field,value,expected'.split(','), (
            (DateField(), datetime.date(2013, 11, 24), '2013-11-24'),
    ))
    def test_datefield_as_string(self, field, value, expected):
        assert field.as_string(value) == expected

    # TimeField ###############################################################

    def test_timefield_1(self):
        f = TimeField(assume_local=False)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.time(18, 43, tzinfo=utc) == f.clean('18:43:00.000Z')
        assert datetime.time(18, 43, tzinfo=utc) == f.clean(datetime.time(18, 43, tzinfo=utc))

    def test_timefield_2(self):
        f = TimeField(assume_local=False, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.time(18, 43, tzinfo=utc) == f.clean('18:43:00.000Z')
        assert datetime.time(18, 43, tzinfo=utc) == f.clean(datetime.time(18, 43, tzinfo=utc))

    @pytest.mark.parametrize('field,value,expected'.split(','), (
            (TimeField(), datetime.time(12, 44, 12, 12), '12:44:12.000012'),
            (TimeField(), datetime.time(12, 44, 12, 12, utc), '12:44:12.000012+00:00'),
    ))
    def test_timefield_as_string(self, field, value, expected):
        assert field.as_string(value) == expected

    # NaiveTimeField ##########################################################

    @pytest.mark.parametrize(('target', 'value', 'expected'), (
            (NaiveTimeField(ignore_timezone=False), '18:43:00.000Z', datetime.time(18, 43, tzinfo=utc)),
            (NaiveTimeField(ignore_timezone=False), '18:43:00.000Z', datetime.time(18, 43, tzinfo=utc)),
            (NaiveTimeField(ignore_timezone=False), datetime.time(18, 43, tzinfo=utc),
             datetime.time(18, 43, tzinfo=utc)),

            (NaiveTimeField(ignore_timezone=False, null=True), None, None),
            (NaiveTimeField(ignore_timezone=False, null=True), '18:43:00.000Z', datetime.time(18, 43, tzinfo=utc)),
            (NaiveTimeField(ignore_timezone=False, null=True), '18:43:00.000Z', datetime.time(18, 43, tzinfo=utc)),
            (NaiveTimeField(ignore_timezone=False, null=True), datetime.time(18, 43, tzinfo=utc),
             datetime.time(18, 43, tzinfo=utc)),

            (NaiveTimeField(ignore_timezone=True), '18:43:00.000Z', datetime.time(18, 43)),
            (NaiveTimeField(ignore_timezone=True), '18:43:00.000Z', datetime.time(18, 43)),
            (NaiveTimeField(ignore_timezone=True), datetime.time(18, 43, tzinfo=utc), datetime.time(18, 43)),
    ))
    def test_naivetimefield__clean_valid_values(self, target, value, expected):
        assert target.clean(value) == expected

    @pytest.mark.parametrize(('target', 'value'), (
            (NaiveTimeField(ignore_timezone=False), None),
            (NaiveTimeField(ignore_timezone=False), 'abc'),
            (NaiveTimeField(ignore_timezone=False), 123),

            (NaiveTimeField(ignore_timezone=False, null=True), 'abc'),
            (NaiveTimeField(ignore_timezone=False, null=True), 123),

            (NaiveTimeField(ignore_timezone=True), None),
            (NaiveTimeField(ignore_timezone=True), 'abc'),
            (NaiveTimeField(ignore_timezone=True), 123),
    ))
    def test_naivetimefield__clean_invalid_values(self, target, value):
        pytest.raises(ValidationError, target.clean, value)

    @pytest.mark.parametrize(('target', 'value', 'expected'), (
            (NaiveTimeField(ignore_timezone=False), datetime.time(18, 43, tzinfo=utc),
             datetime.time(18, 43, tzinfo=utc)),
            (NaiveTimeField(ignore_timezone=True), datetime.time(18, 43, tzinfo=utc), datetime.time(18, 43)),
    ))
    def test_naivetimefield__prepare(self, target, value, expected):
        assert target.prepare(value) == expected

    # DateTimeField ###########################################################

    def test_datetimefield_1(self):
        f = DateTimeField(assume_local=False)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean('2013-11-24T18:43:00.000Z')
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean('2013-11-24 18:43:00.000Z')
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc))

    def test_datetimefield_2(self):
        f = DateTimeField(assume_local=False, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean('2013-11-24T18:43:00.000Z')
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean('2013-11-24 18:43:00.000Z')
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc))

    @pytest.mark.parametrize('field,value,expected'.split(','), (
            (DateTimeField(), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=FixedTimezone.from_hours_minutes(10)),
             '2013-11-24T18:43:00+10:00'),
            (DateTimeField(), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc), '2013-11-24T18:43:00+00:00'),
    ))
    def test_datetimefield_as_string(self, field, value, expected):
        assert field.as_string(value) == expected

    # NaiveDateTimeField ######################################################

    @pytest.mark.parametrize(('target', 'value', 'expected'), (
            (NaiveDateTimeField(ignore_timezone=False), '2013-11-24T18:43:00.000Z',
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),
            (NaiveDateTimeField(ignore_timezone=False), '2013-11-24 18:43:00.000Z',
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),
            (NaiveDateTimeField(ignore_timezone=False), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),

            (NaiveDateTimeField(ignore_timezone=False, null=True), None, None),
            (NaiveDateTimeField(ignore_timezone=False, null=True), '2013-11-24T18:43:00.000Z',
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),
            (NaiveDateTimeField(ignore_timezone=False, null=True), '2013-11-24 18:43:00.000Z',
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),
            (NaiveDateTimeField(ignore_timezone=False, null=True), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),

            (NaiveDateTimeField(ignore_timezone=True), '2013-11-24T18:43:00.000Z',
             datetime.datetime(2013, 11, 24, 18, 43)),
            (NaiveDateTimeField(ignore_timezone=True), '2013-11-24 18:43:00.000Z',
             datetime.datetime(2013, 11, 24, 18, 43)),
            (NaiveDateTimeField(ignore_timezone=True), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
             datetime.datetime(2013, 11, 24, 18, 43)),
    ))
    def test_naivedatetimefield__clean_valid_values(self, target, value, expected):
        assert target.clean(value) == expected

    @pytest.mark.parametrize(('target', 'value'), (
            (NaiveDateTimeField(ignore_timezone=False), None),
            (NaiveDateTimeField(ignore_timezone=False), 'abc'),
            (NaiveDateTimeField(ignore_timezone=False), 123),

            (NaiveDateTimeField(ignore_timezone=False, null=True), 'abc'),
            (NaiveDateTimeField(ignore_timezone=False, null=True), 123),

            (NaiveDateTimeField(ignore_timezone=True), None),
            (NaiveDateTimeField(ignore_timezone=True), 'abc'),
            (NaiveDateTimeField(ignore_timezone=True), 123),
    ))
    def test_naivedatetimefield__clean_invalid_values(self, target, value):
        pytest.raises(ValidationError, target.clean, value)

    @pytest.mark.parametrize(('target', 'value', 'expected'), (
            (NaiveDateTimeField(ignore_timezone=False), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
             datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)),
            (NaiveDateTimeField(ignore_timezone=True), datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
             datetime.datetime(2013, 11, 24, 18, 43)),
    ))
    def test_naivedatetimefield__prepare(self, target, value, expected):
        assert target.prepare(value) == expected

    # HttpDateTimeField #######################################################

    def test_httpdatetimefield_1(self):
        f = HttpDateTimeField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2012, 8, 29, 17, 12, 58, tzinfo=utc) == f.clean('Wed Aug 29 17:12:58 +0000 2012')
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc))

    def test_httpdatetimefield_2(self):
        f = HttpDateTimeField(null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2012, 8, 29, 17, 12, 58, tzinfo=utc) == f.clean('Wed Aug 29 17:12:58 +0000 2012')
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc))

    # TimeStampField ##########################################################

    def test_timestampfield_1(self):
        f = TimeStampField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 'Wed Aug 29 17:12:58 +0000 2012')
        assert datetime.datetime(1970, 1, 1, 0, 2, 3, tzinfo=utc) == f.clean(123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc))

    def test_timestampfield_2(self):
        f = TimeStampField(null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 'Wed Aug 29 17:12:58 +0000 2012')
        assert datetime.datetime(1970, 1, 1, 0, 2, 3, tzinfo=utc) == f.clean(123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc))

    def test_timestampfield_3(self):
        f = TimeStampField()
        assert f.prepare(None) is None
        assert 123 == f.prepare(datetime.datetime(1970, 1, 1, 0, 2, 3, tzinfo=utc))
        assert 123 == f.prepare(123)
        assert 123 == f.prepare(
            datetime.datetime(1970, 1, 1, 10, 2, 3, tzinfo=FixedTimezone.from_hours_minutes(10)))

    # DictField ###############################################################

    def test_dictfield_1(self):
        f = DictField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert {} == f.clean({})
        assert {'foo': 'bar'} == f.clean({'foo': 'bar'})
        assert f.default == dict

    def test_dictfield_2(self):
        f = DictField(null=True)
        assert None == f.clean(None)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert {'foo': 'bar'} == f.clean({'foo': 'bar'})

    # ArrayField ##############################################################

    def test_arrayfield_1(self):
        f = ArrayField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        assert ['foo', 'bar'], f.clean(['foo' == 'bar'])
        assert ['foo', 'bar', '$', 'eek'], f.clean(['foo', 'bar', '$' == 'eek'])
        assert f.default == list

    def test_arrayfield_2(self):
        f = ArrayField(null=True)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        assert ['foo', 'bar'], f.clean(['foo' == 'bar'])
        assert ['foo', 'bar', '$', 'eek'], f.clean(['foo', 'bar', '$' == 'eek'])

    # TypedListField #########################################################

    def test_typedlistfield_1(self):
        f = TypedListField(IntegerField())
        assert "List<Integer>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        pytest.raises(ValidationError, f.clean, ['foo', 'bar'])
        assert [1, 2, 3], f.clean([1, 2 == 3])
        assert f.default == list

    def test_typedlistfield_2(self):
        f = TypedListField(IntegerField(), null=True)
        assert "List<Integer>" == f.data_type_name(f)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        pytest.raises(ValidationError, f.clean, ['foo', 'bar'])
        assert [1, 2, 3], f.clean([1, 2 == 3])

    def test_typed_list_field_dynamic_type_name(self):
        f = TypedListField(DynamicTypeNameFieldTest(), null=True)
        assert "List<Foo>" == f.data_type_name(f)

    # TypedDictField ##########################################################

    def test_typeddictfield_1(self):
        f = TypedDictField(IntegerField())
        assert "Dict<String, Integer>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, {'foo': 'bar'})
        assert {'foo': 1} == f.clean({'foo': 1})

    def test_typeddictfield_2(self):
        f = TypedDictField(IntegerField(), null=True)
        assert "Dict<String, Integer>" == f.data_type_name(f)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, 'abc')
        pytest.raises(ValidationError, f.clean, 123)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, {'foo': 'bar'})
        assert {'foo': 1} == f.clean({'foo': 1})

    def test_typeddictfield_3(self):
        f = TypedDictField(StringField(), IntegerField(), null=True)
        assert "Dict<Integer, String>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, {'foo': 'bar'})
        assert {1: 'foo'} == f.clean({1: 'foo'})

    def test_typeddictfield_nested_typed_array(self):
        f = TypedDictField(TypedArrayField(StringField()))
        assert "Dict<String, List<String>>" == f.data_type_name(f)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, {'foo': 'bar'})
        assert {'foo': ['bar', 'eek']}, f.clean({'foo': ['bar' == 'eek']})

    def test_typeddictfield_validate(self):
        f = TypedDictField(
            IntegerField(min_value=5),
            StringField(max_length=5, choices=[
                ('foo', 'Foo'),
                ('bad_value', 'Bad Value'),
            ])
        )
        assert "Dict<String, Integer>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, {None: 6})
        pytest.raises(ValidationError, f.clean, {'bad_value': 6})
        pytest.raises(ValidationError, f.clean, {'bar': 6})
        pytest.raises(ValidationError, f.clean, {'foo': None})
        pytest.raises(ValidationError, f.clean, {'foo': 2})

    def test_typed_dict_field_dynamic_type_name(self):
        f = TypedDictField(
            DynamicTypeNameFieldTest(),
            DynamicTypeNameFieldTest(),
        )
        assert "Dict<Foo, Foo>" == f.data_type_name(f)

    @pytest.mark.parametrize('field value actual'.split(), (
        (EmailField(), u'foo@example.company', u'foo@example.company'),

        (IPv4Field(), u'127.0.0.1', u'127.0.0.1'),

        (IPv6Field(), u'::1', u'::1'),
        (IPv6Field(), u'1:2:3:4:5:6:7:8', u'1:2:3:4:5:6:7:8'),

        (IPv46Field(), u'127.0.0.1', u'127.0.0.1'),
        (IPv46Field(), u'::1', u'::1'),
        (IPv46Field(), u'1:2:3:4:5:6:7:8', u'1:2:3:4:5:6:7:8'),
    ))
    def test_valid_values(self, field, value, actual):
        assert field.clean(value) == actual



# -*- coding: utf-8 -*-
import pytest
import datetime
from odin import datetimeutil


class TestTimezones(object):
    def test_utc(self):
        assert datetime.timedelta(0) == datetimeutil.utc.utcoffset(datetime.datetime.now())
        assert 'UTC' == datetimeutil.utc.tzname(datetime.datetime.now())
        assert datetime.timedelta(0) == datetimeutil.utc.dst(datetime.datetime.now())
        assert 'UTC' == str(datetimeutil.utc)
        assert '<timezone: UTC>' == repr(datetimeutil.utc)

    def test_fixed_timezone(self):
        target = datetimeutil.FixedTimezone.from_hours_minutes(10, 30)

        assert datetime.timedelta(hours=10, minutes=30) == target.utcoffset(datetime.datetime.now())
        assert '10:30' == target.tzname(datetime.datetime.now())
        assert datetime.timedelta(0) == target.dst(datetime.datetime.now())
        assert '10:30' == str(target)
        assert "<timezone '10:30' datetime.timedelta(0, 37800)>" == repr(target)

    def test_fixed_timezone_pickling(self):
        import pickle

        target = datetimeutil.FixedTimezone.from_hours_minutes(10, 30)

        data = pickle.dumps(target)
        result = pickle.loads(data)

        assert result == target


class TestParseIsoDateString(object):
    @pytest.mark.parametrize('value', (
        '2014-01-13', '2014-1-13'
    ))
    def test_valid_values(self, value):
        assert datetime.date(2014, 1, 13) == datetimeutil.parse_iso_date_string(value)

    @pytest.mark.parametrize('value', (
        123, '2014/01/13'
    ))
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_iso_date_string, value)


class TestParseIsoTimeString(object):
    @pytest.mark.parametrize(('value', 'expected'), (
        ('23:53:25', datetime.time(23, 53, 25, 0, datetimeutil.utc)),
        ('23:53:25Z', datetime.time(23, 53, 25, 0, datetimeutil.utc)),
        ('23:53:25.432Z', datetime.time(23, 53, 25, 432, datetimeutil.utc)),
        ('23:53:25+10', datetime.time(23, 53, 25, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 0))),
        ('23:53:25.432+10', datetime.time(23, 53, 25, 432, datetimeutil.FixedTimezone.from_hours_minutes(10, 0))),
        ('23:53:25+10:30', datetime.time(23, 53, 25, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 30))),
        ('23:53:25.432-10:30', datetime.time(23, 53, 25, 432,
                                             datetimeutil.FixedTimezone.from_hours_minutes(-10, -30))),
    ))
    def test_valid_values(self, value, expected):
        assert datetimeutil.parse_iso_time_string(value) == expected

    @pytest.mark.parametrize(('value', 'default_tz', 'expected'), (
        ('23:53:25', None, datetime.time(23, 53, 25)),
        ('23:53:25', datetimeutil.IgnoreTimezone, datetime.time(23, 53, 25)),
        ('23:53:25Z', None, datetime.time(23, 53, 25, 0, datetimeutil.utc)),
        ('23:53:25Z', datetimeutil.IgnoreTimezone, datetime.time(23, 53, 25, 0)),
    ))
    def test_valid_values_timezone_handling(self, value, default_tz, expected):
        assert datetimeutil.parse_iso_time_string(value, default_tz) == expected

    def test_value_values_with_alternate_timezone(self):
        assert(
            datetime.time(23, 53, 25, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 0)) ==
            datetimeutil.parse_iso_time_string(
                '23:53:25', default_timezone=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))
        )

    @pytest.mark.parametrize('value', (
        123, '23:53:25EST'
    ))
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_iso_time_string, value)


class TestParseIsoDateTimeString(object):
    @pytest.mark.parametrize(('value', 'expected'), (
        ('2014-01-13T00:28:33', datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc)),
        ('2014-01-13T00:28:33Z', datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc)),
        ('2014-01-13T00:28:33.432Z', datetime.datetime(2014, 1, 13, 0, 28, 33, 432, datetimeutil.utc)),
        ('2014-01-13T00:28:33+10', datetime.datetime(2014, 1, 13, 0, 28, 33, 0,
                                                     datetimeutil.FixedTimezone.from_hours_minutes(10, 0))),
        ('2014-01-13T00:28:33.432+10', datetime.datetime(2014, 1, 13, 0, 28, 33, 432,
                                                         datetimeutil.FixedTimezone.from_hours_minutes(10, 0))),
        ('2014-01-13T00:28:33+10:30', datetime.datetime(2014, 1, 13, 0, 28, 33, 0,
                                                        datetimeutil.FixedTimezone.from_hours_minutes(10, 30))),
        ('2014-01-13T00:28:33.432-10:30', datetime.datetime(2014, 1, 13, 0, 28, 33, 432,
                                                            datetimeutil.FixedTimezone.from_hours_minutes(-10, -30))),
        ('2014-01-13T00:28:33.432+1030', datetime.datetime(2014, 1, 13, 0, 28, 33, 432,
                                                           datetimeutil.FixedTimezone.from_hours_minutes(10, 30))),
        # Non-Standard but close formats that are common
        ('2014-01-13 00:28:33Z', datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc)),
        ('2014-01-13 00:28:33 GMT', datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc))
    ))
    def test_valid_values(self, value, expected):
        assert datetimeutil.parse_iso_datetime_string(value) == expected

    @pytest.mark.parametrize(('value', 'default_tz', 'expected'), (
        ('2014-01-13T00:28:33', None, datetime.datetime(2014, 1, 13, 0, 28, 33)),
        ('2014-01-13T00:28:33', datetimeutil.IgnoreTimezone, datetime.datetime(2014, 1, 13, 0, 28, 33)),
        ('2014-01-13T00:28:33Z', None, datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc)),
        ('2014-01-13T00:28:33Z', datetimeutil.IgnoreTimezone, datetime.datetime(2014, 1, 13, 0, 28, 33)),
    ))
    def test_valid_values_timezone_handling(self, value, default_tz, expected):
        assert datetimeutil.parse_iso_datetime_string(value, default_tz) == expected

    @pytest.mark.parametrize('value', (
        123,
        '2014-1-13T00:28:33Z',
        '2014/1/13T00:28:33Z',
        '2014-01-13T00:28:33EST',
    ))
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_iso_datetime_string, value)


class TestToDateString(object):
    def test_naive_datetime(self):
        dt = datetime.datetime(2013, 7, 13, 16, 54, 46, 123000)
        actual = datetimeutil.to_ecma_datetime_string(dt, datetimeutil.utc)  # assume UTC to simplify this particular test
        assert "2013-07-13T16:54:46.123Z" == actual

    def test_aware_datetime(self):
        dt = datetime.datetime(2013, 7, 13, 16, 54, 46, 123000, datetimeutil.utc)
        actual = datetimeutil.to_ecma_datetime_string(dt, datetimeutil.utc)
        assert "2013-07-13T16:54:46.123Z" == actual


class TestParseHttpDateString(object):
    @pytest.mark.parametrize(('value', 'expected'), (
        ('Wed Aug 29 17:12:58 +1030 2012', datetime.datetime(2012, 8, 29, 17, 12, 58, 0,
                                                             datetimeutil.FixedTimezone.from_hours_minutes(10, 30))),
        ('Wed Aug 29 17:12:58 -0900 2012', datetime.datetime(2012, 8, 29, 17, 12, 58, 0,
                                                             datetimeutil.FixedTimezone.from_hours_minutes(-9, 0))),
        ('Wed Aug 29 17:12:58 -0930 2012', datetime.datetime(2012, 8, 29, 17, 12, 58, 0,
                                                             datetimeutil.FixedTimezone.from_hours_minutes(-9, -30))),
        ('Wed Aug 29 17:12:58 +0800 2012', datetime.datetime(2012, 8, 29, 17, 12, 58, 0,
                                                             datetimeutil.FixedTimezone.from_hours_minutes(8, 00))),
        # Alternate position for the year
        ('Wed, Aug 29 2012 17:12:58 -0930', datetime.datetime(2012, 8, 29, 17, 12, 58, 0,
                                                              datetimeutil.FixedTimezone.from_hours_minutes(-9, -30))),
    ))
    def test_valid_values(self, value, expected):
        assert datetimeutil.parse_http_datetime_string(value) == expected

    @pytest.mark.parametrize('value', (
        123,
        '2014-1-13T00:28:33Z',
        '2014/1/13T00:28:33Z',
        '2014-01-13T00:28:33EST',
    ))
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_http_datetime_string, value)


class TestToHttpDateString(object):
    @pytest.mark.parametrize(('value', 'expected'), (
        (datetime.datetime(2012, 8, 29, 17, 12, 58, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 30)),
         'Wed, 29 Aug 2012 06:42:58 GMT'),
        (datetime.datetime(2012, 8, 29, 17, 12, 58, 0, datetimeutil.utc),
         'Wed, 29 Aug 2012 17:12:58 GMT'),
    ))
    def test_valid_values(self, value, expected):
        assert datetimeutil.to_http_datetime_string(value) == expected



import pytest

from odin import datastructures


class TestCaseLessStringList(object):
    @pytest.mark.parametrize('needle haystack'.split(), (
        ('foo', ['foo']),
        ('FOO', ['foo']),
        ('foo', ['FOO']),
        ('Foo', ['fOO']),
        ('Foo', ['foo', 'bar']),
    ))
    def test_contains__true(self, needle, haystack):
        target = datastructures.CaseLessStringList(haystack)
        assert needle in target

    @pytest.mark.parametrize('needle haystack'.split(), (
        ('foo', ['food']),
        ('FOO', ['food']),
        ('foo', ['FOOD']),
        ('Foo', ['fOOd']),
        ('Foo', ['food', 'bar']),
    ))
    def test_contains__false(self, needle, haystack):
        target = datastructures.CaseLessStringList(haystack)
        assert needle not in target

    def test_extend(self):
        target = datastructures.CaseLessStringList(('foo', ))
        target.extend('bar')
        assert target == ['FOO', 'B', 'A', 'R']

    def test_append(self):
        target = datastructures.CaseLessStringList('foo')
        target.append('bar')
        assert 'Bar' in target

    def test_insert(self):
        target = datastructures.CaseLessStringList(('foo', 'bar'))
        target.insert(1, 'eek')
        assert 'Eek' in target
        assert target == ['FOO', 'EEK', 'BAR']

    def test_index(self):
        target = datastructures.CaseLessStringList(('foo', 'bar', 'eek'))

        assert target.index('Foo') == 0
        assert target.index('Bar') == 1
        assert target.index('EEk', 1) == 2

    def test_remove(self):
        target = datastructures.CaseLessStringList(('foo', 'bar', 'eek'))
        target.remove('Bar')
        assert target == ['FOO', 'EEK']



class ResourceIterable(object):
    """
    Iterable object that yields resources.
    """
    def __iter__(self):
        raise NotImplementedError()


class TypedResourceIterable(ResourceIterable):
    """
    Iterable object that yields a specific resource.
    """
    def __init__(self, resource_type):
        self.resource_type = resource_type

    def __iter__(self):
        raise NotImplementedError()



"""
This module is to include utils for managing compatibility between Python and Odin releases.
"""
import inspect
import warnings


def deprecated(message, category=DeprecationWarning):
    """
    Decorator for marking classes/functions as being deprecated and are to be removed in the future.

    :param message: Message provided.
    :param category: Category of warning, defaults to DeprecationWarning

    """
    def wrap(obj):
        if inspect.isclass(obj):
            old_init = obj.__init__

            def wrapped_init(*args, **kwargs):
                warnings.warn(
                    "{0} is deprecated and scheduled for removal. {1}".format(obj.__name__, message),
                    category=category
                )
                return old_init(*args, **kwargs)

            obj.__init__ = wrapped_init
            return obj

        else:
            def wrapped_func(*args, **kwargs):
                warnings.warn(
                    "{0} is deprecated and scheduled for removal. {1}".format(obj.__name__, message),
                    category=category
                )
                return obj(*args, **kwargs)

            return wrapped_func
    return wrap



# -*- coding: utf-8 -*-
import datetime
import re
import time
import six
import sys
from email.utils import parsedate_tz as parse_http_datetime
from email.utils import formatdate as format_http_datetime  # noqa


class IgnoreTimezone(object):
    pass


ZERO = datetime.timedelta(0)
LOCAL_STD_OFFSET = datetime.timedelta(seconds=-time.timezone)
LOCAL_DST_OFFSET = datetime.timedelta(seconds=-time.altzone) if time.daylight else LOCAL_STD_OFFSET
LOCAL_DST_DIFF = LOCAL_DST_OFFSET - LOCAL_STD_OFFSET


class UTC(datetime.tzinfo):
    """
    UTC timezone.
    """
    def utcoffset(self, _):
        return ZERO

    def tzname(self, _):
        return "UTC"

    def dst(self, _):
        return ZERO

    def __str__(self):
        return "UTC"

    def __repr__(self):
        return "<timezone: %s>" % self

utc = UTC()


class LocalTimezone(datetime.tzinfo):
    """
    The current local timezone (according to the platform)
    """
    def utcoffset(self, dt):
        return LOCAL_DST_OFFSET if self._is_dst(dt) else LOCAL_STD_OFFSET

    def dst(self, dt):
        return LOCAL_DST_DIFF if self._is_dst(dt) else ZERO

    def tzname(self, dt):
        return time.tzname[self._is_dst(dt)]

    def _is_dst(self, dt):
        stamp = time.mktime((dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.weekday(), 0, 0))
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0

    def __str__(self):
        return time.tzname[0]

    def __repr__(self):
        return "<timezone: %s>" % self

local = LocalTimezone()


class FixedTimezone(datetime.tzinfo):
    """
    A fixed timezone for when a timezone is specified by a numerical offset and no dst information is available.
    """
    __slots__ = ('offset', 'name',)

    @classmethod
    def from_seconds(cls, seconds):
        sign = '-' if seconds < 0 else ''
        minutes = abs(seconds // 60)
        hours = minutes // 60
        minutes %= 60
        name = "%s%02d:%02d" % (sign, hours, minutes)

        if sign == '-':
            hours *= -1
            minutes *= -1

        return cls(datetime.timedelta(hours=hours, minutes=minutes), name)

    @classmethod
    def from_hours_minutes(cls, hours, minutes=0):
        sign = '-' if hours < 0 else ''
        hours = abs(hours)
        minutes = abs(minutes)
        name = "%s%02d:%02d" % (sign, hours, minutes)

        if sign == '-':
            hours *= -1
            minutes *= -1

        return cls(datetime.timedelta(hours=hours, minutes=minutes), name)

    @classmethod
    def from_groups(cls, groups, default_timezone=utc):
        tz = groups['timezone']
        if tz is None:
            return default_timezone

        if tz in ('Z', 'GMT', 'UTC'):
            return utc

        sign = groups['tz_sign']
        hours = int(groups['tz_hour'])
        minutes = int(groups['tz_minute'] or 0)
        name = "%s%02d:%02d" % (sign, hours, minutes)

        if sign == '-':
            hours = -hours
            minutes = -minutes

        return cls(datetime.timedelta(hours=hours, minutes=minutes), name)

    def __init__(self, offset=None, name=None):
        super(FixedTimezone, self).__init__()
        self.offset = offset or ZERO
        self.name = name or ''

    def utcoffset(self, _):
        return self.offset

    def dst(self, _):
        return ZERO

    def tzname(self, _):
        return self.name

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<timezone %r %r>" % (self.name, self.offset)

    def __eq__(self, other):
        return self.offset == other.offset

    # Pickle support

    def __getstate__(self):
        return {
            'offset': self.offset,
            'name': self.name,
        }

    def __setstate__(self, state):
        self.offset = state.get('offset')
        self.name = state.get('name')


def get_tz_aware_dt(dt, assumed_tz=local):
    """
    Get a time zone aware date time from a supplied date time.

    If dt is already timezone aware it will be returned unchanged.
    If dt is not aware it will be assumed that dt is in local time.
    """
    assert isinstance(dt, datetime.datetime)

    if dt.tzinfo:
        return dt
    else:
        return dt.replace(tzinfo=assumed_tz)


def now_utc():
    """
    Get now in UTC (with timezone set correctly).
    """
    return datetime.datetime.now(tz=utc)


def now_local():
    """
    Get now in the current local timezone.
    """
    return datetime.datetime.now(tz=local)


# Fallback for python 2.6
if sys.version_info[0] == 2 and sys.version_info[1] == 6:
    def total_seconds(timedelta):
        """
        Backported implementation of total_seconds
        """
        return (
            timedelta.microseconds + 0.0 +
            (timedelta.seconds + timedelta.days * 24 * 3600)
        )
else:
    total_seconds = datetime.timedelta.total_seconds


UNIX_EPOCH = datetime.datetime(1970, 1, 1, tzinfo=utc)
if six.PY3:
    to_timestamp = datetime.datetime.timestamp
else:
    def to_timestamp(dt):
        return total_seconds((dt - UNIX_EPOCH))


ISO8601_TIME_STRING_RE = re.compile(
    r"^(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})(\.(?P<microseconds>\d+))?"
    r"(?P<timezone>Z|((?P<tz_sign>[-+])(?P<tz_hour>\d{2})(:(?P<tz_minute>\d{2}))?))?$")
ISO8601_DATETIME_STRING_RE = re.compile(
    r"^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})[tT\s]"
    r"(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})(\.(?P<microseconds>\d+))? ?"
    r"(?P<timezone>Z|GMT|UTC|((?P<tz_sign>[-+])(?P<tz_hour>\d{2})(:?(?P<tz_minute>\d{2}))?))?$")


def parse_iso_date_string(date_string):
    """
    Parse a date in the string format defined in ISO 8601.
    """
    if not isinstance(date_string, six.string_types):
        raise ValueError("Expected string")

    try:
        return datetime.datetime.strptime(date_string, '%Y-%m-%d').date()
    except ValueError:
        raise ValueError("Expected ISO 8601 formatted date string.")


def parse_iso_time_string(time_string, default_timezone=utc):
    """
    Parse a time in the string format defined by ISO 8601.
    """
    if not isinstance(time_string, six.string_types):
        raise ValueError("Expected string")

    matches = ISO8601_TIME_STRING_RE.match(time_string)
    if not matches:
        raise ValueError("Expected ISO 8601 formatted time string.")

    groups = matches.groupdict()
    if default_timezone is IgnoreTimezone:
        tz = None
    else:
        tz = FixedTimezone.from_groups(groups, default_timezone)
    return datetime.time(
        int(groups['hour']),
        int(groups['minute']),
        int(groups['second']),
        int(groups['microseconds'] or 0),
        tz
    )


def parse_iso_datetime_string(datetime_string, default_timezone=utc):
    """
    Parse a datetime in the string format defined by ISO 8601.
    """
    if not isinstance(datetime_string, six.string_types):
        raise ValueError("Expected string")

    matches = ISO8601_DATETIME_STRING_RE.match(datetime_string)
    if not matches:
        raise ValueError("Expected ISO 8601 formatted datetime string.")

    groups = matches.groupdict()
    if default_timezone is IgnoreTimezone:
        tz = None
    else:
        tz = FixedTimezone.from_groups(groups, default_timezone)
    return datetime.datetime(
        int(groups['year']),
        int(groups['month']),
        int(groups['day']),
        int(groups['hour']),
        int(groups['minute']),
        int(groups['second']),
        int(groups['microseconds'] or 0),
        tz
    )


def to_ecma_datetime_string(dt, default_timezone=local):
    """
    Convert a python datetime into the string format defined by ECMA-262.

    See ECMA international standard: ECMA-262 section 15.9.1.15

    ``assume_local_time`` if true will assume the date time is in local time if the object is a naive date time object;
        else assumes the time value is utc.
    """
    assert isinstance(dt, datetime.datetime)

    dt = get_tz_aware_dt(dt, default_timezone).astimezone(utc)
    return "%4i-%02i-%02iT%02i:%02i:%02i.%03iZ" % (
        dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond / 1000)


def parse_http_datetime_string(datetime_string):
    """
    Parse a datetime in the string format defined by ISO-1123 (or HTTP date time).
    """
    elements = None
    if isinstance(datetime_string, six.string_types):
        elements = parse_http_datetime(datetime_string)

    if not elements:
        raise ValueError("Expected ISO-1123 formatted datetime string.")

    return datetime.datetime(*elements[:6], tzinfo=FixedTimezone.from_seconds(elements[-1]))


def to_http_datetime_string(dt, default_timezone=local):
    """
    Convert a python datetime into the string format defined by ISO-1123 (or HTTP date time).
    """
    assert isinstance(dt, datetime.datetime)

    dt = get_tz_aware_dt(dt, default_timezone).astimezone(utc)
    timeval = time.mktime(dt.timetuple())
    now = time.localtime(timeval)
    return '{0}, {1:02d} {2} {3:04d} {4:02d}:{5:02d}:{6:02d} {7}'.format(
        ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'][now[6]],
        now[2],
        ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
         'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'][now[1] - 1],
        now[0], now[3], now[4], now[5], 'GMT')



"""
Tools for generating Swagger Specification from resources.
"""
from odin.utils import getmeta
from odin import fields

SWAGGER_SPEC_TYPE_MAPPING = {
    fields.IntegerField: 'integer',
    fields.FloatField: 'number',
    fields.BooleanField: 'boolean',
}
"""
Mapping of fields to Swagger types.
"""

SWAGGER_SPEC_FORMAT_MAPPING = {
    fields.StringField: '',
    fields.IntegerField: 'int64',
    fields.FloatField: 'float',
    fields.BooleanField: '',
    fields.DateField: 'date',
    fields.DateTimeField: 'date-time',
    fields.NaiveTimeField: 'date-time',
}
"""
Mapping of fields to Swagger formats.
"""


def generate_definition(resource):
    """
    Generate a `Swagger Definitions Object <http://swagger.io/specification/#definitionsObject>`_
    from a resource.

    """
    meta = getmeta(resource)

    definition = {
        'type': "object",
        'properties': {}
    }

    for field in meta.all_fields:
        field_definition = {
            'type': SWAGGER_SPEC_TYPE_MAPPING.get(field, 'string')
        }

        if field in SWAGGER_SPEC_FORMAT_MAPPING:
            field_definition['format'] = SWAGGER_SPEC_FORMAT_MAPPING[field]

        if field.doc_text:
            field_definition['description'] = field.doc_text

        definition['properties'][field.name] = field_definition

    return {meta.name: definition}



# -*- coding: utf-8 -*-
import pytest
from odin import utils
from .resources import Book


class TestCamelAndDashFormats(object):
    def test_camel_to_lower_separated(self):
        assert "this_is_an_example" == utils.camel_to_lower_underscore('thisIsAnExample')
        assert "this_is_an_example" == utils.camel_to_lower_underscore('ThisIsAnExample')
        assert "this-is-a-test" == utils.camel_to_lower_dash('thisIsATest')

    def test_lower_underscore_to_camel(self):
        assert "thisIsAnExample" == utils.lower_underscore_to_camel('this_is_an_example')
        assert "thisIsAnExample" == utils.lower_underscore_to_camel('this_is_an_example')
        assert "thisIsATest" == utils.lower_underscore_to_camel('this_Is_A_test')

    def test_lower_dash_to_camel(self):
        assert "thisIsAnExample" == utils.lower_dash_to_camel('this-is-an-example')
        assert "thisIsAnExample" == utils.lower_dash_to_camel('this-is-an-example')
        assert "thisIsATest" == utils.lower_dash_to_camel('this-Is-A-test')


class TestIterator(object):
    def test_extract_fields_from_dict(self):
        source_dict = {
            'title': "My Title",
            'num_pages': 1,
            'rrp': 1.2,
            'fiction': True,
            'published': ['no_validation_for_type'],
            'another_field': 'that is not on the resource'
        }

        assert {
                   'title': "My Title",
                   'num_pages': 1,
                   'rrp': 1.2,
                   'fiction': True,
                   'published': ['no_validation_for_type'],
               } == utils.extract_fields_from_dict(source_dict, Book)


class TestValueInChoice(object):
    CHOICES = (
        ('foo', 'Foo'),
        ('bar', 'Foo'),
        (123, 'Number'),
        (True, 'Boolean'),
        (None, 'None'),
        ('eek', 'With Help', 'Help string'),
    )

    def test_value_in_choices(self):
        assert utils.value_in_choices('foo', self.CHOICES)
        assert utils.value_in_choices(123, self.CHOICES)
        assert utils.value_in_choices(True, self.CHOICES)
        assert utils.value_in_choices(None, self.CHOICES)
        assert utils.value_in_choices('eek', self.CHOICES)
        assert not utils.value_in_choices('xyz', self.CHOICES)
        assert not utils.value_in_choices(False, self.CHOICES)
        assert not utils.value_in_choices(321, self.CHOICES)


class TestForceTuple(object):
    """
    Tests for odin.utils.force_tuple
    """

    @pytest.mark.parametrize(('value', 'expected'), (
            (None, tuple()),
            ('', ('',)),
            (0, (0,)),
            (False, (False,)),
            (True, (True,)),
            (123, (123,)),
            ('Foo', ('Foo',)),
            (['Foo', 'bar'], ('Foo', 'bar')),
            (('Foo', 'bar'), ('Foo', 'bar')),
            (['Foo', 123, 'bar', 'eek'], ('Foo', 123, 'bar', 'eek')),
    ))
    def test_values(self, value, expected):
        assert utils.force_tuple(value) == expected


class TestChunk(object):
    """
    Tests for odin.utils.chunk
    """

    @pytest.mark.parametrize(('value', 'size', 'expected'), (
        ("This is a block of text split by 4", 4, [
            ['T', 'h', 'i', 's'],
            [' ', 'i', 's', ' '],
            ['a', ' ', 'b', 'l'],
            ['o', 'c', 'k', ' '],
            ['o', 'f', ' ', 't'],
            ['e', 'x', 't', ' '],
            ['s', 'p', 'l', 'i'],
            ['t', ' ', 'b', 'y'],
            [' ', '4']
        ]),
    ))
    def test_small_values(self, value, size, expected):
        """Small scale test (eg small iterable that can be easily verified)"""
        assert [list(c) for c in utils.chunk(value, size)] == expected



# -*- coding: utf-8 -*-
import datetime
from odin import datetimeutil
from odin import serializers


class TestSerializer(object):
    def test_date_iso_format(self):
        assert "2014-01-16" == serializers.date_iso_format(datetime.date(2014, 1, 16))

    def test_time_iso_format(self):
        assert ("00:29:32+00:00" ==
                serializers.TimeIsoFormat(datetimeutil.utc)(datetime.time(0, 29, 32)))
        assert ("00:29:32+10:00" ==
                serializers.TimeIsoFormat(datetimeutil.utc)(datetime.time(0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))))
        assert ("00:29:32-03:30" ==
                serializers.TimeIsoFormat(datetimeutil.FixedTimezone.from_hours_minutes(-3, -30))(datetime.time(0, 29, 32)))
        assert ("00:29:32+10:00" ==
                serializers.TimeIsoFormat(datetimeutil.FixedTimezone.from_hours_minutes(-3, -30))(
                datetime.time(0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))))

    def test_datetime_iso_format(self):
        assert ("2014-01-16T00:29:32+00:00" ==
                serializers.DatetimeIsoFormat(datetimeutil.utc)(datetime.datetime(2014, 1, 16, 0, 29, 32)))
        assert ("2014-01-16T00:29:32+10:00" ==
                serializers.DatetimeIsoFormat(datetimeutil.utc)(datetime.datetime(2014, 1, 16, 0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))))

        assert ("2014-01-16T00:29:32-03:30" ==
                serializers.DatetimeIsoFormat(datetimeutil.FixedTimezone.from_hours_minutes(-3, -30))(datetime.datetime(2014, 1, 16, 0, 29, 32)))
        assert ("2014-01-16T00:29:32+10:00" ==
                serializers.DatetimeIsoFormat(datetimeutil.FixedTimezone.from_hours_minutes(-3, -30))(
                datetime.datetime(2014, 1, 16, 0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))))

    def test_datetime_ecma_format(self):
        assert ("2014-01-16T00:29:32.000Z" ==
                serializers.DatetimeEcmaFormat(False, default_timezone=datetimeutil.utc)(datetime.datetime(2014, 1, 16, 0, 29, 32)))
        assert ("2014-01-15T14:29:32.000Z" ==
                serializers.DatetimeEcmaFormat(False, default_timezone=datetimeutil.utc)(
                datetime.datetime(2014, 1, 16, 0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))))

        assert ("2014-01-16T03:59:32.000Z" ==
                serializers.DatetimeEcmaFormat(False, default_timezone=datetimeutil.FixedTimezone.from_hours_minutes(-3, -30))(
                datetime.datetime(2014, 1, 16, 0, 29, 32)))
        assert ("2014-01-15T14:29:32.000Z" ==
                serializers.DatetimeEcmaFormat(False, default_timezone=datetimeutil.FixedTimezone.from_hours_minutes(-3, -30))(
                datetime.datetime(2014, 1, 16, 0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0))))




# -*- coding: utf-8 -*-
import pytest
from odin.contrib.money import AmountField
from odin.contrib.money import Amount
from odin.exceptions import ValidationError

a = Amount(11)
b = Amount(22, "AUD")
c = Amount(33, "AUD")
d = Amount(44, "NZD")


class TestAmountFields(object):
    # AmountField #############################################################

    def test_money_field(self):
        f = AmountField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'a')
        assert a == f.clean(11)
        assert b == f.clean((22, 'AUD'))

    def test_money_field_1(self):
        f = AmountField(allowed_currencies=('NZD', 'AUD'))
        assert d == f.clean((44, 'NZD'))
        pytest.raises(ValidationError, f.clean, (22, 'USD'))



# -*- coding: utf-8 -*-
import os
import pytest
import odin
import odin.exceptions
from odin.codecs import csv_codec

FIXTURE_PATH_ROOT = os.path.join(os.path.dirname(__file__), "fixtures")


class Book(odin.Resource):
    title = odin.StringField(name="Title")
    num_pages = odin.IntegerField(name="Num Pages")
    rrp = odin.FloatField(name="RRP")
    genre = odin.StringField(name="Genre", choices=(
        ('sci-fi', 'Science Fiction'),
        ('fantasy', 'Fantasy'),
        ('others', 'Others'),
    ), null=True)
    author = odin.StringField(name="Author")
    publisher = odin.StringField(name="Publisher")
    language = odin.StringField(name="Language", null=True)

    def extra_attrs(self, attrs):
        self.extras = attrs


class TestCsvResourceReader(object):
    def test_valid(self):
        with open(os.path.join(FIXTURE_PATH_ROOT, "library-valid.csv")) as f:
            books = [book for book in csv_codec.ResourceReader(f, Book)]

        assert 6 == len(books)
        assert "Consider Phlebas" == books[0].title

    def test_invalid(self):
        with open(os.path.join(FIXTURE_PATH_ROOT, "library-invalid.csv")) as f:
            with pytest.raises(odin.exceptions.ValidationError):
                books = [book for book in csv_codec.ResourceReader(f, Book)]


class TestReader(object):
    def open_fixture(self, file_name):
        return open(os.path.join(FIXTURE_PATH_ROOT, file_name))

    @pytest.mark.parametrize('fixture options'.split(), (
        ('library-valid.csv', dict(includes_header=True)),
        ('library-header-alt-order.csv', dict(includes_header=True)),
        ('library-last-bad.csv', dict(includes_header=True)),
        ('library-no-header.csv', dict(includes_header=False)),
        ('library-lower-header.csv', dict(includes_header=True,
                                          ignore_header_case=True)),
        ('library-valid.csv', dict(includes_header=True, strict_fields=True)),
    ))
    def test_valid_libraries(self, fixture, options):
        with self.open_fixture(fixture) as f:
            target = csv_codec.reader(f, Book, **options)
            library = list(target)
        assert len(library) == 6
        assert [book.title for book in library] == [
            'Consider Phlebas',
            'The Moonstone',
            'Casino Royale',
            'A Clockwork Orange',
            'The Naked God',
            'Equal Rites',
        ]
        assert [book.num_pages for book in library] == [
            471, 462, 181, 139, 1256, 282,
        ]
        assert [book.rrp for book in library] == [
            19.5, 17.95, 9.95, 9.95, 19.95, 16.95,
        ]
        genres = [book.genre for book in library]
        assert (
            genres == ['sci-fi', 'others', 'others', 'others', 'sci-fi', 'fantasy']
            or
            genres == [None, None, None, None, None, None]
        )
        assert [book.author for book in library] == [
            'Iain M. Banks',
            'Wilkie Collins',
            'Ian Fleming',
            'Anthony Burgess',
            'Peter F. Hamilton',
            'Terry Pratchett',
        ]
        assert [book.publisher for book in library] == [
            'Macmillan',
            'Harper & Row',
            'Penguin Books',
            'Penguin Books',
            'Macmillan',
            'Corgi Books',
        ]
        assert [book.language for book in library] == [
            'English', '', 'English', '', 'English', 'English'
        ]

    def test_extra_fields(self):
        with self.open_fixture('library-header-alt-order.csv') as f:
            target = csv_codec.reader(f, Book, includes_header=True)
            library = list(target)

        assert all(book.extras == ['Other field'] for book in library)

    def test_headers_not_strict(self):
        with self.open_fixture('library-header-alt-order.csv') as f:
            with pytest.raises(odin.exceptions.CodecDecodeError) as result:
                csv_codec.reader(f, Book, includes_header=True, strict_fields=True)

            assert str(result.value) == 'Extra unknown fields: Else'

    def test_default_empty_value_is_none(self):
        with self.open_fixture('library-valid.csv') as f:
            target = csv_codec.reader(f, Book, includes_header=True)
            target.default_empty_value = None
            library = list(target)

        assert [book.language for book in library] == [
            'English', None, 'English', None, 'English', 'English'
        ]

    def test_error(self):
        with self.open_fixture('library-invalid.csv') as f:
            target = csv_codec.reader(f, Book, includes_header=True)

            with pytest.raises(odin.exceptions.ValidationError):
                list(target)

    def test_error_handler(self):
        errors = []

        def error_handler(_, idx):
            errors.append(idx)

        with self.open_fixture('library-invalid.csv') as f:
            target = csv_codec.reader(f, Book, includes_header=True, error_callback=error_handler)
            library = list(target)

        assert len(library) == 4
        assert len(errors) == 2
        assert errors == [3, 5]

    def test_error_handler_returns_false(self):
        with self.open_fixture('library-invalid.csv') as f:
            target = csv_codec.reader(f, Book, includes_header=True, error_callback=lambda v, i: False)

            with pytest.raises(odin.exceptions.ValidationError):
                list(target)



# -*- coding: utf-8 -*-
from odin.utils import getmeta

try:
    import msgpack
except ImportError:
    raise ImportError("odin.codecs.msgpack_codec requires the 'msgpack-python' package.")  # noqa

import datetime

from odin import bases
from odin import serializers, resources, ResourceAdapter


TYPE_SERIALIZERS = {
    datetime.date: serializers.date_iso_format,
    datetime.time: serializers.time_iso_format,
    datetime.datetime: serializers.datetime_iso_format,
}
CONTENT_TYPE = 'application/x-msgpack'


class OdinPacker(msgpack.Packer):
    """
    Encoder for Odin resources.
    """
    def __init__(self, include_virtual_fields=True, *args, **kwargs):
        kwargs.setdefault('default', self.default)
        super(OdinPacker, self).__init__(*args, **kwargs)
        self.include_virtual_fields = include_virtual_fields

    def default(self, o):
        if isinstance(o, (resources.ResourceBase, ResourceAdapter)):
            meta = getmeta(o)
            obj = o.to_dict(self.include_virtual_fields)
            obj[meta.type_field] = meta.resource_name
            return obj
        elif isinstance(o, bases.ResourceIterable):
            return list(o)
        elif o.__class__ in TYPE_SERIALIZERS:
            return TYPE_SERIALIZERS[o.__class__](o)


def load(fp, resource=None, encoding='UTF8', full_clean=True, default_to_not_supplied=False):
    """
    Load a from a MessagePack encoded file.

    See :py:meth:`loads` for more details of the loading operation.

    :param fp: a file pointer to read MessagePack data from.
    :param resource: A resource instance or a resource name to use as the base for creating a resource.
    :param encoding: Encoding to use when loading file
    :param full_clean: Do a full clean of the object as part of the loading process.
    :returns: A resource object or object graph of resources loaded from file.
    """
    return resources.build_object_graph(msgpack.load(fp, encoding=encoding), resource, full_clean,
                                        default_to_not_supplied)


def loads(s, resource=None, encoding='UTF8', full_clean=True, default_to_not_supplied=False):
    """
    Load from a MessagePack encoded string/bytes.

    If a ``resource`` value is supplied it is used as the base resource for the supplied MessagePack data. I one is not
    supplied a resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError``
    will be raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in
    the file to be a child object from within the inheritance tree.

    :param s: String to load and parse.
    :param resource: A resource instance or a resource name to use as the base for creating a resource.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param encoding: Encoding to use when loading file
    :returns: A resource object or object graph of resources parsed from supplied string.
    """
    return resources.build_object_graph(msgpack.loads(s, encoding=encoding), resource, full_clean, False,
                                        default_to_not_supplied)


def dump(resource, fp, cls=OdinPacker, include_virtual_fields=True, **kwargs):
    """
    Dump to a MessagePack encoded file.

    :param resource: The root resource to dump to a MessagePack encoded file.
    :param cls: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :param fp: The file pointer that represents the output file.
    """
    fp.write(cls(include_virtual_fields, **kwargs).pack(resource))


def dumps(resource, cls=OdinPacker, include_virtual_fields=True, **kwargs):
    """
    Dump to a MessagePack encoded string.

    :param resource: The root resource to dump to a MessagePack encoded file.
    :param cls: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :returns: MessagePack encoded string.
    """
    return cls(include_virtual_fields, **kwargs).pack(resource)



import six
from odin.utils import getmeta

from odin import bases
from odin import resources, ResourceAdapter


TYPE_SERIALIZERS = {}


class OdinEncoder(object):
    def __init__(self, include_virtual_fields=True, include_type_field=True):
        self.include_virtual_fields = include_virtual_fields
        self.include_type_field = include_type_field

    def default(self, o):
        if isinstance(o, (resources.ResourceBase, ResourceAdapter)):
            meta = getmeta(o)
            obj = o.to_dict(self.include_virtual_fields)
            if self.include_type_field:
                obj[meta.type_field] = meta.resource_name
            return obj
        elif isinstance(o, bases.ResourceIterable):
            return list(o)
        elif o.__class__ in TYPE_SERIALIZERS:
            return TYPE_SERIALIZERS[o.__class__](o)
        return o

    def encode(self, o):
        _encoder = _make_encoder(self.default)
        return _encoder(o)


load = resources.build_object_graph
load.__doc__ = """
    Load a :py:class:`dict` into a Resource structure.

    This method is an alias of :py:func:`odin.

    :param d: Dict to load

    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :raises ValidationError: During building of the object graph and issues discovered are raised as a ValidationError.

    """


def dump(resource, cls=OdinEncoder, **kwargs):
    """
    Dump a resource structure into a nested :py:class:`dict`.

    While a resource includes a *to_dict* method this method is not recursive. The dict codec recursively iterates
    through the resource structure to produce a full dict. This is useful for testing for example.

    :param resource: The root resource to dump
    :param cls: Encoder class to utilise
    :return:

    """
    encoder = cls(**kwargs)
    return encoder.encode(resource)


def _make_encoder(_default):
    def _encode_list(lst):
        return [_encode(o) for o in lst]

    def _encode_dict(dct):
        return dict((k, _encode(o)) for k, o in six.iteritems(dct))

    def _encode(o):
        if isinstance(o, (list, tuple)):
            return _encode_list(o)
        elif isinstance(o, dict):
            return _encode_dict(o)
        else:
            o = _default(o)
            if isinstance(o, (list, tuple, dict)):
                return _encode(o)
            return o
    return _encode



# -*- coding: utf-8 -*-
import logging

from odin.utils import getmeta

logger = logging.Logger('odin.registration')


def generate_mapping_cache_name(from_obj, to_obj):
    return "%s.%s > %s.%s" % (from_obj.__module__, from_obj.__name__, to_obj.__module__, to_obj.__name__,)


class ResourceCache(object):
    # Use the Borg pattern to share state between all instances. Details at
    # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66531.
    __shared_state = dict(
        resources={},
        mappings={},
        field_resolvers=set(),
        validation_error_handlers={},
    )

    def __init__(self):
        self.__dict__ = self.__shared_state

    def register_resources(self, *resources):
        """
        Register a resource (or resources).

        :param resources: Argument list of resources to register.

        """
        for resource in resources:
            meta = getmeta(resource)
            resource_name = meta.resource_name.lower()
            self.resources[resource_name] = resource
            class_name = meta.class_name.lower()
            if resource_name != class_name:
                self.resources[class_name] = resource

            logger.debug("Registered resource <%s>", class_name)

    def get_resource(self, resource_name):
        """
        Get a resource by name.

        :param resource_name: Name of the resource to find.
        :returns: The resource type that matches requested name (case insensitive); or :const:`None` if the requested
            name has not been registered.

        """
        return self.resources.get(resource_name.lower())

    def register_mapping(self, mapping):
        """
        Register a mapping

        :param mapping: Mapping object to register.

        """
        mapping_name = generate_mapping_cache_name(mapping.from_obj, mapping.to_obj)
        self.mappings[mapping_name] = mapping

        logger.debug("Registered mapping <%s>", mapping_name)

    def get_mapping(self, from_obj, to_obj):
        """
        Get a mapping based on the from and to objects (likely to be resources).

        :param from_obj: Object to map from.
        :param to_obj: Object to map to.
        :returns: A mapping object that supports mapping from *from_obj* to *to_obj*
        :raises: KeyError if a mapping cannot be found.

        """
        mapping_name = generate_mapping_cache_name(from_obj, to_obj)
        return self.mappings[mapping_name]

    def register_field_resolver(self, resolver, base_type):
        """
        Register a field resolver.

        The *base_type* will also cover all subclasses.

        :param base_type: Base type for subclasses that this resolver will work with.
        :param resolver: Resolver object used to resolve subclasses of *base_type*.

        """
        self.field_resolvers.add((base_type, resolver))

        logger.info("Registered field resolver <%s>", resolver)

    def get_field_resolver(self, obj_type):
        """
        Get a field resolver for an object type.

        :param obj_type: Object type to find a field resolver for.
        :return: A field resolver instance for resolving fields on *obj_type*.
        :raises: KeyError if a resolver cannot be found.

        """
        for base_type, field_resolver in self.field_resolvers:
            if issubclass(obj_type, base_type):
                return field_resolver(obj_type)
        raise KeyError('No field resolver could be found for %r' % obj_type)

    def register_validation_error_handler(self, error_type, handler):
        """
        Register a validation error handler.

        :param handler: A method that can handle the exception type.
        :param error_type: Error exception to register a handler for.

        """
        self.validation_error_handlers[error_type] = handler

    def get_validation_error_list(self):
        """
        Get a list of validation errors that can be used in an exception clause.

        :return: List of error types.

        """
        try:
            return self._validation_error_list
        except AttributeError:
            validation_error_list = tuple(self.validation_error_handlers.keys())
            self._validation_error_list = validation_error_list
            return validation_error_list

    def get_validation_error_handler(self, error_type):
        """
        Get the handler for a particular error_type.

        :param error_type: Error exception related to a handler.
        :return: Handler

        """
        return self.validation_error_handlers[error_type.__class__]


cache = ResourceCache()

register_resources = cache.register_resources
get_resource = cache.get_resource

register_mapping = cache.register_mapping
get_mapping = cache.get_mapping

register_field_resolver = cache.register_field_resolver
get_field_resolver = cache.get_field_resolver

register_validation_error_handler = cache.register_validation_error_handler
get_validation_error_list = cache.get_validation_error_list
get_validation_error_handler = cache.get_validation_error_handler



# -*- coding: utf-8 -*-
"""
CSV Codec
~~~~~~~~~

Codec for iterating a CSV file and parsing into a Resource.

The CSV codec is codec that yields multiple resources rather than a single document.
The CSV codec does not support nesting of resources.

Reading data from a CSV file::

    with open("my_file.csv") as f:
        with resource in csv_codec.reader(f, MyResource):
            ...

"""
import six
import csv

from odin import bases
from odin.compatibility import deprecated
from odin.datastructures import CaseLessStringList
from odin.fields import NOT_PROVIDED
from odin.resources import create_resource_from_iter, create_resource_from_dict
from odin.utils import getmeta, lazy_property
from odin.exceptions import CodecDecodeError, ValidationError

CONTENT_TYPE = 'text/csv'


class Reader(bases.TypedResourceIterable):
    """
    Customisable reader object.
    """
    csv_reader = csv.reader
    """
    CSV Reader object to use (if you wish to use *unicodecsv* or similar)
    """

    includes_header = True
    """
    File is expected to include a header.
    """

    ignore_header_case = False
    """
    Use case-less comparison on header fields.
    """

    strict_fields = False
    """
    Strictly check header fields.
    """

    csv_dialect = 'excel'
    """
    CSV Dialect to use; defaults to the CSV libraries default value of *excel*.
    """

    default_empty_value = ''
    """
    The default value to use if a field is empty. This can be used to default to *None*.
    """

    def __init__(self, f, resource_type, full_clean=True, error_callback=None, **reader_kwargs):
        """
        Initialise a reader

        :param f: Input file (or file like) object to read
        :param resource_type: Resource type to use as field template.
        :param full_clean: Perform a full clean on objects
        :param error_callback: Optional callback for errors
        :param reader_kwargs: kwargs to pass to the csv_reader

        """
        super(Reader, self).__init__(resource_type)
        self.full_clean = full_clean
        if error_callback:
            self.handle_validation_error = error_callback

        # Backwards compatibility
        for arg in ('csv_reader', 'includes_header', 'ignore_header_case', 'strict_fields', 'csv_dialect'):
            if arg in reader_kwargs:
                setattr(self, arg, reader_kwargs.pop(arg))

        # Create reader instance
        self._reader = self._create_reader(f, reader_kwargs)

        # Configure header
        if self.includes_header:
            self.header = self._read_header()

            # Handle strict fields
            if self.strict_fields and self.extra_field_names:
                raise CodecDecodeError("Extra unknown fields: {0}".format(','.join(self.extra_field_names)))

        # Built in counters
        self.row_count = None
        self.error_count = None

    def __iter__(self):
        # Reset error count
        self.error_count = 0

        # Local vars
        resource = self.resource_type
        full_clean = self.full_clean
        default_empty_value = self.default_empty_value
        handle_validation_error = getattr(self, 'handle_validation_error', None)
        idx = -1

        def create_resource(values, i):
            try:
                return create_resource_from_iter(
                    # Handle empty values
                    (default_empty_value if v == '' else v for v in values),
                    resource, full_clean
                )
            except ValidationError as ve:
                # Don't raise these through yield as will cause a StopIteration
                # even if validation error can be handled safely.
                self.error_count += 1
                if not handle_validation_error:
                    raise
                # If handle error explicitly returns False raise exception
                if handle_validation_error(ve, i) is False:
                    raise

        if self.includes_header:
            mapping = self.field_mapping

            for idx, row in enumerate(self._reader):
                # Check if row is less than mapping (as this will causes errors)!
                res = create_resource(
                    (s if s is NOT_PROVIDED else row[s] for s in mapping),
                    idx + 1)  # Add one to index as row "0" will be the header
                if res:
                    yield res
        else:
            for idx, row in enumerate(self._reader):
                res = create_resource(row, idx)
                if res:
                    yield res

        self.row_count = idx + 1  # Add one to get a count from the last index

    def _create_reader(self, f, kwargs):
        """
        Create internal reader instance

        :param f: File (or file like) object
        :param kwargs: Dictionary of additional keyword args
        :return: Reader instance

        """
        return self.csv_reader(f, self.csv_dialect, **kwargs)

    def _read_header(self):
        """
        Get the header, this needs to be called **once** only!
        """
        header = next(self._reader)
        if self.ignore_header_case:
            header = CaseLessStringList(header)
        return header

    @lazy_property
    def field_names(self):
        """
        Field names from resource.
        """
        fields = getmeta(self.resource_type).fields
        if self.ignore_header_case:
            return CaseLessStringList(field.name for field in fields)
        else:
            return tuple(field.name for field in fields)

    @lazy_property
    def extra_field_names(self):
        """
        Extra fields not included in header
        """
        return tuple(field for field in self.header if field not in self.field_names)

    @lazy_property
    def field_mapping(self):
        """
        Index mapping of CSV fields to resource fields.
        """
        mapping = []

        # Add expected fields
        header = self.header
        for name in self.field_names:
            if name in header:
                mapping.append(header.index(name))
            else:
                mapping.append(NOT_PROVIDED)

        # Append any extra fields
        for name in self.extra_field_names:
            mapping.append(header.index(name))

        return tuple(mapping)


def reader(f, resource, includes_header=False, csv_module=csv, full_clean=True,
           ignore_header_case=False, strict_fields=False, **kwargs):
    """
    CSV reader that returns resource objects

    :param f: file like object
    :param resource:
    :param includes_header: File includes a header that should be used to map columns
    :param csv_module: Specify an alternate csv module (eg unicodecsv); defaults to the builtin csv as this module
        is implemented in C.
    :param full_clean: Perform a full clean on each object
    :param ignore_header_case: Ignore the letter case on header
    :param strict_fields: Extra fields cannot be provided.
    :return: Iterable reader object
    :rtype: Reader

    """
    return Reader(f, resource, full_clean,
                  csv_reader=csv_module.reader,
                  includes_header=includes_header,
                  ignore_header_case=ignore_header_case,
                  strict_fields=strict_fields,
                  **kwargs)


@deprecated("This class will be removed in 1.1 migrate to the `reader` method.")
class ResourceReader(csv.DictReader):
    def __init__(self, f, resource, *args, **kwargs):
        self.resource = resource
        csv.DictReader.__init__(self, f, *args, **kwargs)

    # Python 2
    def next(self):
        return create_resource_from_dict(csv.DictReader.next(self), self.resource, copy_dict=False)

    # Python 3
    def __next__(self):
        return create_resource_from_dict(csv.DictReader.__next__(self), self.resource, copy_dict=False)


def value_fields(resource):
    """
    Iterator to get non-composite (eg value) fields for export
    """
    meta = getmeta(resource)
    return [f for f in meta.all_fields if f not in meta.composite_fields]


def _get_resource_type(resources, resource_type):
    if isinstance(resources, bases.TypedResourceIterable):
        # Use first resource to obtain field list
        return resource_type or resources.resource_type
    elif isinstance(resources, bases.ResourceIterable) and resource_type:
        return resource_type
    elif isinstance(resources, (list, tuple)):
        if not len(resources):
            return
        # Use first resource to obtain field list
        return resource_type or resources[0]
    else:
        raise Exception("Not supported input format")


def dump_to_writer(writer, resources, resource_type=None, fields=None):
    """
    Dump resources to a CSV writer interface.

    The interface should expose the :py:class:`csv.writer` interface.

    :type writer: :py:class:`csv.writer`
    :param writer: Writer object
    :param fields: List of fields to write
    :param resources: Collection of resources to dump.
    :param resource_type: Resource type to use for CSV columns; if None the first resource will be used.
    :returns: List of fields that where written to.

    """
    resource_type = resource_type or _get_resource_type(resources, resource_type)

    if not fields:
        fields = value_fields(resource_type)

    # Iterate resources and write to CSV
    for resource in resources:
        row = [field.prepare(field.value_from_object(resource)) for field in fields]
        writer.writerow(row)

    return fields


def dump(f, resources, resource_type=None, include_header=True, cls=csv.writer, **kwargs):
    """
    Dump resources into a CSV file.

    :param f: File to dump to.
    :param resources: Collection of resources to dump.
    :param resource_type: Resource type to use for CSV columns; if None the first resource will be used.
    :param include_header: Write a CSV header.
    :param cls: Writer to use when writing CSV, this should be based on :class:`csv.writer`.
    :param kwargs: Additional parameters to be supplied to the writer instance.

    """
    resource_type = _get_resource_type(resources, resource_type)

    fields = value_fields(resource_type)

    # Setup CSV
    writer = cls(f, **kwargs)

    # Write header
    if include_header:
        writer.writerow([field.name for field in fields])

    dump_to_writer(writer, resources, resource_type, fields)


def dumps(resources, resource_type=None, cls=csv.writer, **kwargs):
    """
    Dump output to a string

    :param resources:
    :param resources: Collection of resources to dump.
    :param resource_type: Resource type to use for CSV columns; if None the first resource will be used.
    :param cls: Writer to use when writing CSV, this should be based on :class:`csv.writer`.
    :param kwargs: Additional parameters to be supplied to the writer instance.

    """
    buf = six.StringIO.StringIO()
    dump(buf, resources, resource_type, cls, **kwargs)
    return buf.getvalue()



"""
Odin Data-structures
~~~~~~~~~~~~~~~~~~~~

Custom data-structures used within odin.

"""


class CaseLessStringList(list):
    """
    List that assumes all elements are lists and performs case less comparisons on them.

    Like a standard list this structure makes a copy of the existing data, while converting all entries to upper case
    strings.

    .. note:: All items added to this list are converted to upper case

    """
    def __init__(self, iterable):
        super(CaseLessStringList, self).__init__(str(i).upper() for i in iterable)

    def __contains__(self, item):
        return super(CaseLessStringList, self).__contains__(str(item).upper())

    def extend(self, iterable):
        return super(CaseLessStringList, self).extend(str(i).upper() for i in iterable)

    def append(self, p_object):
        return super(CaseLessStringList, self).append(str(p_object).upper())

    def insert(self, index, p_object):
        return super(CaseLessStringList, self).insert(index, str(p_object).upper())

    def index(self, value, *args, **kwargs):
        return super(CaseLessStringList, self).index(str(value).upper(), *args, **kwargs)

    def remove(self, value):
        return super(CaseLessStringList, self).remove(str(value).upper())



from __future__ import absolute_import, print_function
from humanfriendly import tables
from odin.fields import NOT_PROVIDED
from odin.utils import field_iter
from .base import SummaryBase


class ResourceSummary(SummaryBase):
    """
    Produce a text based summary of a resource
    """
    def __init__(self, resource, out):
        super(ResourceSummary, self).__init__(out)
        self.resource = resource

    def render(self):
        data = []
        column_names = ('Name', 'Type', 'Default', 'Doc Text')
        for field in field_iter(self.resource):
            data.append([
                field.name,
                field.__class__.__name__,
                '' if field is not NOT_PROVIDED else str(field.default),
                field.doc_text
            ])
        self.print(tables.format_pretty_table(data, column_names))



# -*- coding: utf-8 -*-
"""
Support for geographic data types.

Latitude, Longitude and LatLng values.
"""
from __future__ import absolute_import
from .datatypes import *  # NoQA
from .fields import *  # NoQA



# -*- coding: utf-8 -*-
import pytest
import six
from odin.contrib.geo import datatypes


class TestDataTypes(object):
    def test_to_dms(self):
        # I am aware that round in python is not accurate but in this case I have calculated the test values
        # specifically so that the rounding will not effect the outcome of this test.
        d, m, s = datatypes.to_dms(27.3375)
        assert (27, 20, 15) == (d, m, round(s, 5))

        d, m, s = datatypes.to_dms(-27.3375)
        assert (-27, 20, 15) == (d, m, round(s, 5))

        d, m, s = datatypes.to_dms(-27.3375, True)
        assert (27, 20, 15) == (d, m, round(s, 5))

    def test_to_dm(self):
        # I am aware that round in python is not accurate but in this case I have calculated the test values
        # specifically so that the rounding will not effect the outcome of this test.
        d, m = datatypes.to_dm(27.3375)
        assert (27, 20.25) == (d, round(m, 5))

        d, m = datatypes.to_dm(-27.3375)
        assert (-27, 20.25) == (d, round(m, 5))

        d, m = datatypes.to_dm(-27.3375, True)
        assert (27, 20.25) == (d, round(m, 5))

    def test_latitude_valid(self):
        assert 30 == datatypes.latitude(30)
        assert -30 == datatypes.latitude(-30)
        assert 90.0 == datatypes.latitude(90.0)
        assert -90.0 == datatypes.latitude(-90.0)
        assert 42.0 == datatypes.latitude("42.0")
        assert -42.0 == datatypes.latitude("-42.0")
        assert 42.0 == datatypes.latitude(float(42.0))
        assert -42.0 == datatypes.latitude(float(-42.0))
        assert 42.0 == datatypes.latitude(datatypes.latitude(42.0))
        assert -42.0 == datatypes.latitude(datatypes.latitude(-42.0))

    def test_latitude_invalid(self):
        pytest.raises(TypeError, datatypes.latitude, None)
        pytest.raises(ValueError, datatypes.latitude, 'a')
        pytest.raises(ValueError, datatypes.latitude, 92.0)
        pytest.raises(ValueError, datatypes.latitude, -92.0)
        pytest.raises(TypeError, datatypes.latitude, 1, 2)

    def test_latitude_str(self):
        lat = datatypes.latitude(27.3375)
        assert "27.3375" == repr(lat)
        assert u"27°20'15.000000\"N" == six.text_type(lat)

        lat = datatypes.latitude(-27.3375)
        assert "-27.3375" == repr(lat)
        assert u"27°20'15.000000\"S" == six.text_type(lat)

    def test_longitude_valid(self):
        assert 30 == datatypes.longitude(30)
        assert -30 == datatypes.longitude(-30)
        assert 180.0 == datatypes.longitude(180.0)
        assert -180.0 == datatypes.longitude(-180.0)
        assert 42.0 == datatypes.longitude("42.0")
        assert -42.0 == datatypes.longitude("-42.0")
        assert 42.0 == datatypes.longitude(float(42.0))
        assert -42.0 == datatypes.longitude(float(-42.0))
        assert 42.0 == datatypes.longitude(datatypes.longitude(42.0))
        assert -42.0 == datatypes.longitude(datatypes.longitude(-42.0))

    def test_longitude_invalid(self):
        pytest.raises(TypeError, datatypes.longitude, None)
        pytest.raises(ValueError, datatypes.longitude, 'a')
        pytest.raises(ValueError, datatypes.longitude, 182.0)
        pytest.raises(ValueError, datatypes.longitude, -182.0)
        pytest.raises(TypeError, datatypes.longitude, 1, 2)

    def test_longitude_str(self):
        lat = datatypes.longitude(27.3375)
        assert "27.3375" == repr(lat)
        assert u"027°20'15.000000\"E" == six.text_type(lat)

        lat = datatypes.longitude(-27.3375)
        assert "-27.3375" == repr(lat)
        assert u"027°20'15.000000\"W" == six.text_type(lat)

    def test_latlng_valid(self):
        assert (10.0, 20.0) == datatypes.latlng(10, 20)
        assert (10.0, 20.0) == datatypes.latlng('10', '20')
        assert (10.0, 20.0) == datatypes.latlng(datatypes.latitude(10), datatypes.longitude(20))
        assert (10.0, 20.0) == datatypes.latlng((10, 20))
        assert (90.0, 180.0) == datatypes.latlng(90, 180)
        assert (-90.0, -180.0) == datatypes.latlng(-90, -180)

    def test_latlng_invalid(self):
        pytest.raises(TypeError, datatypes.latlng, None)
        pytest.raises(TypeError, datatypes.latlng, 1)
        pytest.raises(TypeError, datatypes.latlng, 1, 2, 3)
        pytest.raises(TypeError, datatypes.latlng, (1, 2, 3))
        pytest.raises(ValueError, datatypes.latlng, 90, 'a')
        pytest.raises(ValueError, datatypes.latlng, 100, 100)

    def test_latlng_str(self):
        ll = datatypes.latlng(27.3375, -27.3375)
        assert "latlng(27.3375, -27.3375)" == repr(ll)
        assert u"(27°20'15.000000\"N, 027°20'15.000000\"W)" == six.text_type(ll)

    def test_latlng_properties(self):
        ll = datatypes.latlng(27.3375, -27.3375)
        assert 27.3375 == ll.lat
        assert -27.3375 == ll.lng

    def test_point_valid(self):
        assert (1, 2) == datatypes.point(1, 2)
        assert (1, 2) == datatypes.point('1', '2')
        assert (1, 2) == datatypes.point((1, 2))
        assert (1, 2, 3) == datatypes.point(1, 2, 3)
        assert (1, 2, 3) == datatypes.point('1', '2', '3')
        assert (1, 2, 3) == datatypes.point((1, 2, 3))

    def test_point_invalid(self):
        pytest.raises(TypeError, datatypes.point, None)
        pytest.raises(TypeError, datatypes.point, 1)
        pytest.raises(TypeError, datatypes.point, 1, 2, 3, 4)
        pytest.raises(TypeError, datatypes.point, (1, 2, 3, 4))
        pytest.raises(ValueError, datatypes.point, 90, 'a', 1)

    def test_point_str(self):
        p = datatypes.point(1, 2)
        assert "point(1.000000, 2.000000)" == repr(p)
        assert "(1.000000, 2.000000)" == str(p)

        p = datatypes.point(1, 2, 3)
        assert "point(1.000000, 2.000000, 3.000000)" == repr(p)
        assert "(1.000000, 2.000000, 3.000000)" == str(p)

    def test_point_properties(self):
        p = datatypes.point(1, 2)
        assert 1 == p.x
        assert 2 == p.y
        pytest.raises(AttributeError, lambda: p.z)
        assert not p.is_3d

        p = datatypes.point(1, 2, 3)
        assert 1 == p.x
        assert 2 == p.y
        assert 3 == p.z
        assert p.is_3d



# -*- coding: utf-8 -*-
import pytest
from odin.contrib.money import datatypes

a = datatypes.Amount(11)
b = datatypes.Amount(22, "AUD")
c = datatypes.Amount(33, "AUD")
d = datatypes.Amount(44, "NZD")


class TestDataTypes(object):
    def test_amount_init(self):
        assert "10.0000" == str(datatypes.Amount(10))
        assert "10.0000" == str(datatypes.Amount("10"))
        assert "10.00 AUD", str(datatypes.Amount("10" == "AUD"))
        assert "-12.30 AUD", str(datatypes.Amount("-12.3" == "AUD"))
        assert "12.35 USD", str(datatypes.Amount(12.345 == "USD"))
        pytest.raises(TypeError, datatypes.Amount, None)
        pytest.raises(ValueError, datatypes.Amount, "abs")
        pytest.raises(TypeError, datatypes.Amount, 12, object())
        # Unknown currency
        pytest.raises(KeyError, datatypes.Amount, 12, 'ZZZ')
        # From tuple
        assert "10.0000" == str(datatypes.Amount(("10",)))
        assert "10.00 NZD", str(datatypes.Amount(("10" == "NZD")))
        pytest.raises(ValueError, datatypes.Amount, ("10", "NZD", "hmmm"))

    # These tests assume that the decimal library is correct.

    def test_amount_type_conversion(self):
        assert 12 == int(datatypes.Amount(12.345))
        assert 12.34 == float(datatypes.Amount(12.34))
        assert "<Amount: 12.34, <Currency: NZD>>", repr(datatypes.Amount(12.34 == 'NZD'))

    def test_amount_neg_pos(self):
        assert "-11.0000" == str(-a)
        assert "-22.00 AUD" == str(-b)
        assert "11.0000" == str(+a)
        assert "22.00 AUD" == str(+b)

    def test_amount_add(self):
        assert "33.00 AUD" == str(a + b)
        assert "55.00 AUD" == str(b + c)
        assert "55.00 NZD" == str(d + a)
        pytest.raises(ValueError, lambda: c + d)

    def test_amount_sub(self):
        assert "-11.00 AUD" == str(a - b)
        assert "-11.00 AUD" == str(b - c)
        assert "33.00 NZD" == str(d - a)
        pytest.raises(ValueError, lambda: c - d)

    def test_amount_mul(self):
        assert "22.0000" == str(a * 2)
        assert "44.00 AUD" == str(b * 2)
        pytest.raises(TypeError, lambda: d * a)

    def test_amount_div(self):
        assert "11.00 AUD" == str(b / 2)
        assert 1.5 == c / b
        pytest.raises(ValueError, lambda: c / d)

    def test_amount_eq(self):
        assert not a == 11
        assert a + b == c
        assert not b * 2 == d
        # and ne
        assert not a + b != c

    def test_amount_lt(self):
        assert a < b
        assert b < c
        pytest.raises(ValueError, lambda: c < d)
        # and le
        assert a <= b
        assert b <= c
        assert a + b <= c
        pytest.raises(ValueError, lambda: c <= d)

    def test_amount_gt(self):
        assert b > a
        assert c > b
        pytest.raises(ValueError, lambda: d > c)
        # and ge
        assert b >= a
        assert c >= b
        assert c >= a + b
        pytest.raises(ValueError, lambda: d >= c)

    def test_amount_format(self):
        assert "11.00" == a.format("{value_raw:0.2f}")
        assert "$22.00 AUD" == b.format("{currency.symbol}{value} {currency.code}")

    def test_assign_currency(self):
        target = a.assign_currency("NZD")
        assert a.is_naive
        assert target.currency == 'NZD'

        pytest.raises(ValueError, b.assign_currency, 'AUD')



# -*- coding: utf-8 -*-
from six import StringIO
from odin import bases
from odin import resources, ResourceAdapter
from odin.exceptions import CodecEncodeError
from odin.utils import getmeta

try:
    import yaml
except ImportError:
    raise ImportError("odin.codecs.yaml_codec requires the 'pyyaml' package.")  # noqa

try:
    from yaml import CSafeLoader as SafeLoader, CSafeDumper as SafeDumper
except ImportError:
    from yaml import SafeLoader, SafeDumper


YAML_TYPES = {}
CONTENT_TYPE = 'application/x-yaml'


class OdinDumper(SafeDumper):
    def __init__(self, stream, include_virtual_fields=True, include_type_field=True, *args, **kwargs):
        SafeDumper.__init__(self, stream, *args, **kwargs)
        self.include_virtual_fields = include_virtual_fields
        self.include_type_field = include_type_field

    def represent_resource(self, data):
        obj = data.to_dict(self.include_virtual_fields)
        if self.include_type_field:
            meta = getmeta(data)
            obj[meta.type_field] = meta.resource_name
        return self.represent_dict(obj)


OdinDumper.add_multi_representer(resources.ResourceBase, OdinDumper.represent_resource)
OdinDumper.add_multi_representer(ResourceAdapter, OdinDumper.represent_resource)
OdinDumper.add_multi_representer(bases.ResourceIterable, OdinDumper.represent_list)


def load(fp, resource=None, full_clean=True, default_to_not_supplied=False):
    """
    Load a from a YAML encoded file.

    If a ``resource`` value is supplied it is used as the base resource for the supplied YAML. I one is not supplied a
    resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError`` will be
    raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in the file
    to be a child object from within the inheritance tree.

    :param fp: a file pointer to read YAML data from.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources loaded from file.

    """
    # try:
    return resources.build_object_graph(yaml.load(fp, SafeLoader), resource, full_clean, False,
                                        default_to_not_supplied)
    # except (ValueError, TypeError) as ex:
    #     raise CodecDecodeError(str(ex))

loads = load


def dump(resource, fp, dumper=OdinDumper, **kwargs):
    """
    Dump to a JSON encoded file.

    :param resource: The root resource to dump to a JSON encoded file.
    :param dumper: Dumper to use serializing to a string; default is the :py:class:`OdinDumper`.
    :param fp: The file pointer that represents the output file.

    """
    try:
        yaml.dump(resource, fp, Dumper=dumper, **kwargs)
    except ValueError as ex:
        raise CodecEncodeError(str(ex))


def dumps(resource, dumper=OdinDumper, **kwargs):
    """
    Dump to a JSON encoded string.

    :param resource: The root resource to dump to a JSON encoded file.
    :param dumper: Dumper to use serializing to a string; default is the :py:class:`OdinDumper`.
    :returns: YAML encoded string.

    """
    stream = StringIO()
    dump(resource, stream, dumper, **kwargs)
    return stream.getvalue()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import six
from io import StringIO
from xml.sax import saxutils
from odin import serializers
from odin import fields
from odin.fields import composite
from odin.utils import attribute_field_iter_items, element_field_iter_items, getmeta

XML_TYPES = {
    datetime.date: serializers.date_iso_format,
    datetime.time: serializers.time_iso_format,
    datetime.datetime: serializers.datetime_iso_format,
}
if not six.PY3:
    XML_TYPES[unicode] = lambda v: v  # noqa

CONTENT_TYPE = 'application/xml'


# class OdinContentHandler(sax.ContentHandler):
#     def __init__(self, resource):
#         self.elements = []
#         self.resources = []
#         self.resource = resource
#
#     def startDocument(self):
#         print("startDocument")
#         self.elements = []
#         self.resources = []
#
#     def endDocument(self):
#         print("endDocument")
#
#     def startElement(self, name, attrs):
#         print("startElement", name, attrs['name'] if 'name' in attrs else '')
#
#         self.elements.append(name)
#
#     def endElement(self, name):
#         print("endElement", name)
#
#         self.elements.pop()
#
#     def startElementNS(self, name, qname, attrs):
#         print("startElementNS", name, qname, attrs)
#
#     def endElementNS(self, name, qname):
#         print("endElementNS", name, qname)
#
#     def characters(self, content):
#         print("characters", content)
#
#     def processingInstruction(self, target, data):
#         print("processingInstruction", target, data)
#
#     def ignorableWhitespace(self, whitespace):
#         print("ignorableWhitespace", whitespace)
#
#     def skippedEntity(self, name):
#         print("skippedEntity", name)
#
#     def startPrefixMapping(self, prefix, uri):
#         print("startPrefixMapping", prefix, uri)
#
#     def endPrefixMapping(self, prefix):
#         print("endPrefixMapping", prefix)
#
#     def setDocumentLocator(self, locator):
#         print("setDocumentLocator", locator)
#
#
# def load(fp, resource=None):
#     handler = OdinContentHandler(resource)
#     sax.parse(fp, handler)
#
#
# def loads(s, resource=None):
#     handler = OdinContentHandler(resource)
#     sax.parseString(s, handler)


def _serialize_to_string(value):
    if value.__class__ in XML_TYPES:
        return XML_TYPES[value.__class__](value)
    else:
        return str(value)


def dump(fp, resource, line_ending=''):
    """
    Dump a resource to a file like object.
    :param fp: File pointer or file like object.
    :param resource: Resource to dump
    :param line_ending:
    """
    meta = getmeta(resource)

    # Write container and any attributes
    attributes = ''.join(
        " %s=%s" % (f.name, saxutils.quoteattr(_serialize_to_string(v)))  # Encode attributes
        for f, v in attribute_field_iter_items(resource)
    )
    fp.write("<%s%s>%s" % (meta.name, attributes, line_ending))

    # Write any element fields
    for field, value in element_field_iter_items(resource):
        if isinstance(field, composite.ListOf):
            if field.use_container:
                fp.write("<%s>%s" % (field.name, line_ending))
            for v in value:
                dump(fp, v, line_ending)
            if field.use_container:
                fp.write("</%s>%s" % (field.name, line_ending))

        elif isinstance(field, composite.DictAs):
            if value is not None:
                dump(fp, value, line_ending)

        elif isinstance(field, fields.ArrayField):
            for v in value:
                fp.write("<%s>%s</%s>%s" % (field.name, _serialize_to_string(v), field.name, line_ending))

        else:
            fp.write("<%s>%s</%s>%s" %
                     (field.name, saxutils.escape(_serialize_to_string(value)), field.name, line_ending))

    fp.write("</%s>%s" % (meta.name, line_ending))


def dumps(resource, **kwargs):
    """
    Dump a resource to a string.

    :param resource: Resource to dump
    :return:
    """
    f = StringIO()
    dump(f, resource, **kwargs)
    return f.getvalue()
